import datetime as dt
from unittest.mock import patch

import pytz

from django.utils.timezone import now
from esi.models import Token
from eveuniverse.models import EveMarketPrice, EveMoon, EveType

from app_utils.testdata_factories import UserFactory
from app_utils.testing import NoSocketsTestCase

from ..constants import EveTypeId
from ..core import CalculatedExtraction, CalculatedExtractionProduct
from ..models import (
    EveOreType,
    Extraction,
    NotificationType,
    OreQualityClass,
    OreRarityClass,
    Refinery,
)
from . import helpers
from .testdata.esi_client_stub import esi_client_stub
from .testdata.factories import (
    CalculatedExtractionFactory,
    ExtractionFactory,
    ExtractionProductFactory,
    MiningLedgerRecordFactory,
    MoonFactory,
    MoonProductFactory,
    NotificationFactory,
    NotificationFactory2,
    OwnerFactory,
    RefineryFactory,
)
from .testdata.load_allianceauth import load_allianceauth
from .testdata.load_eveuniverse import load_eveuniverse, nearest_celestial_stub

MODELS_PATH = "moonmining.models"


class TestEveOreTypeCalcRefinedValues(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eveuniverse()

    def setUp(self) -> None:
        self.cinnebar = EveOreType.objects.get(id=45506)
        tungsten = EveType.objects.get(id=16637)
        mercury = EveType.objects.get(id=16646)
        evaporite_deposits = EveType.objects.get(id=16635)
        EveMarketPrice.objects.create(eve_type=tungsten, average_price=7000)
        EveMarketPrice.objects.create(eve_type=mercury, average_price=9750)
        EveMarketPrice.objects.create(eve_type=evaporite_deposits, average_price=950)

    def test_should_return_value_per_unit(self):
        self.assertEqual(self.cinnebar.calc_refined_value_per_unit(0.7), 4002.25)


class TestEveOreTypeProfileUrl(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eveuniverse()

    def test_should_return_correct_value(self):
        # given
        cinnebar = EveOreType.objects.get(id=45506)
        # when
        result = cinnebar.profile_url
        # then
        self.assertEqual(result, "https://www.kalkoken.org/apps/eveitems/?typeId=45506")


# class TestExtractionIsJackpot(NoSocketsTestCase):
#     @classmethod
#     def setUpClass(cls):
#         super().setUpClass()
#         load_eveuniverse()
#         load_allianceauth()
#         moon = helpers.create_moon_40161708()
#         owner = Owner.objects.create(
#             corporation=EveCorporationInfo.objects.get(corporation_id=2001)
#         )
#         cls.refinery = Refinery.objects.create(
#             id=40161708, moon=moon, owner=owner, eve_type_id=35835
#         )
#         cls.ore_quality_regular = EveOreType.objects.get(id=45490)
#         cls.ore_quality_improved = EveOreType.objects.get(id=46280)
#         cls.ore_quality_excellent = EveOreType.objects.get(id=46281)
#         cls.ore_quality_excellent_2 = EveOreType.objects.get(id=46283)

#     def test_should_be_jackpot(self):
#         # given
#         extraction = ExtractionFactory(
#             refinery=self.refinery,
#             chunk_arrival_at=now() + dt.timedelta(days=3),
#             auto_fracture_at=now() + dt.timedelta(days=4),
#             started_at=now() - dt.timedelta(days=3),
#             status=Extraction.Status.STARTED,
#         )
#         ExtractionProductFactory(
#             extraction=extraction,
#             ore_type=self.ore_quality_excellent,
#             volume=1000000 * 0.1,
#         )
#         ExtractionProductFactory(
#             extraction=extraction,
#             ore_type=self.ore_quality_excellent_2,
#             volume=1000000 * 0.1,
#         )
#         # when
#         result = extraction.calc_is_jackpot()
#         # then
#         self.assertTrue(result)

#     def test_should_not_be_jackpot_1(self):
#         # given
#         extraction = ExtractionFactory(
#             refinery=self.refinery,
#             chunk_arrival_at=now() + dt.timedelta(days=3),
#             auto_fracture_at=now() + dt.timedelta(days=4),
#             started_at=now() - dt.timedelta(days=3),
#             status=Extraction.Status.STARTED,
#         )
#         ExtractionProductFactory(
#             extraction=extraction,
#             ore_type=self.ore_quality_excellent,
#             volume=1000000 * 0.1,
#         )
#         ExtractionProductFactory(
#             extraction=extraction,
#             ore_type=self.ore_quality_improved,
#             volume=1000000 * 0.1,
#         )
#         # when
#         result = extraction.calc_is_jackpot()
#         # then
#         self.assertFalse(result)

#     def test_should_not_be_jackpot_2(self):
#         # given
#         extraction = ExtractionFactory(
#             refinery=self.refinery,
#             chunk_arrival_at=now() + dt.timedelta(days=3),
#             auto_fracture_at=now() + dt.timedelta(days=4),
#             started_at=now() - dt.timedelta(days=3),
#             status=Extraction.Status.STARTED,
#         )
#         ExtractionProductFactory(
#             extraction=extraction,
#             ore_type=self.ore_quality_improved,
#             volume=1000000 * 0.1,
#         )
#         ExtractionProductFactory(
#             extraction=extraction,
#             ore_type=self.ore_quality_excellent,
#             volume=1000000 * 0.1,
#         )
#         # when
#         result = extraction.calc_is_jackpot()
#         # then
#         self.assertFalse(result)

#     def test_should_not_be_jackpot_3(self):
#         # given
#         extraction = ExtractionFactory(
#             refinery=self.refinery,
#             chunk_arrival_at=now() + dt.timedelta(days=3),
#             auto_fracture_at=now() + dt.timedelta(days=4),
#             started_at=now() - dt.timedelta(days=3),
#             status=Extraction.Status.STARTED,
#         )
#         ExtractionProductFactory(
#             extraction=extraction,
#             ore_type=self.ore_quality_regular,
#             volume=1000000 * 0.1,
#         )
#         ExtractionProductFactory(
#             extraction=extraction,
#             ore_type=self.ore_quality_improved,
#             volume=1000000 * 0.1,
#         )
#         # when
#         result = extraction.calc_is_jackpot()
#         # then
#         self.assertFalse(result)

#     def test_should_not_be_jackpot_4(self):
#         # given
#         extraction = ExtractionFactory(
#             refinery=self.refinery,
#             chunk_arrival_at=now() + dt.timedelta(days=3),
#             auto_fracture_at=now() + dt.timedelta(days=4),
#             started_at=now() - dt.timedelta(days=3),
#             status=Extraction.Status.STARTED,
#         )
#         # when
#         result = extraction.calc_is_jackpot()
#         # then
#         self.assertFalse(result)


class TestExtraction(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eveuniverse()
        load_allianceauth()

    def test_should_convert_to_calculated_extraction(self):
        # given
        refinery = RefineryFactory()
        my_map = [
            (Extraction.Status.STARTED, CalculatedExtraction.Status.STARTED),
            (Extraction.Status.CANCELED, CalculatedExtraction.Status.CANCELED),
            (Extraction.Status.READY, CalculatedExtraction.Status.READY),
            (Extraction.Status.COMPLETED, CalculatedExtraction.Status.COMPLETED),
        ]
        for in_status, out_status in my_map:
            with self.subTest(status=in_status):
                extraction = ExtractionFactory(status=in_status, refinery=refinery)
                # when
                obj = extraction.to_calculated_extraction()
                # then
                self.assertEqual(obj.status, out_status)


class TestMoonUpdateValue(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eveuniverse()

    @patch(MODELS_PATH + ".MOONMINING_VOLUME_PER_MONTH", 1000000)
    @patch(MODELS_PATH + ".MOONMINING_REPROCESSING_YIELD", 0.7)
    def test_should_calc_correct_value(self):
        # given
        moon = MoonFactory(create_products=False)
        helpers.generate_market_prices(use_process_pricing=False)
        MoonProductFactory(moon=moon, ore_type_id=EveTypeId.CINNABAR, amount=0.19)
        MoonProductFactory(moon=moon, ore_type_id=EveTypeId.CUBIC_BISTOT, amount=0.23)
        MoonProductFactory(
            moon=moon, ore_type_id=EveTypeId.FLAWLESS_ARKONOR, amount=0.25
        )
        MoonProductFactory(
            moon=moon, ore_type_id=EveTypeId.STABLE_VELDSPAR, amount=0.33
        )
        # when
        result = moon.calc_value()
        # then
        self.assertEqual(result, 84622187.5)

    def test_should_return_zero_if_prices_are_missing(self):
        # given
        moon = MoonFactory()
        # when
        result = moon.calc_value()
        # then
        self.assertEqual(result, 0)


class TestMoonCalcRarityClass(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eveuniverse()
        cls.ore_type_r0 = EveOreType.objects.get(id=EveTypeId.CUBIC_BISTOT)
        cls.ore_type_r4 = EveOreType.objects.get(id=EveTypeId.BITUMENS)
        cls.ore_type_r8 = EveOreType.objects.get(id=EveTypeId.EUXENITE)
        cls.ore_type_r16 = EveOreType.objects.get(id=EveTypeId.CHROMITE)
        cls.ore_type_r32 = EveOreType.objects.get(id=EveTypeId.CINNABAR)
        cls.ore_type_r64 = EveOreType.objects.get(id=EveTypeId.XENOTIME)

    def test_should_return_R4(self):
        # given
        moon = MoonFactory(create_products=False)
        MoonProductFactory(moon=moon, ore_type=self.ore_type_r0, amount=0.23)
        MoonProductFactory(moon=moon, ore_type=self.ore_type_r4, amount=0.19)
        # when
        result = moon.calc_rarity_class()
        # then
        self.assertEqual(result, OreRarityClass.R4)

    def test_should_return_R8(self):
        # given
        moon = MoonFactory(create_products=False)
        MoonProductFactory(moon=moon, ore_type=self.ore_type_r8, amount=0.25)
        MoonProductFactory(moon=moon, ore_type=self.ore_type_r0, amount=0.23)
        MoonProductFactory(moon=moon, ore_type=self.ore_type_r4, amount=0.19)
        # when
        result = moon.calc_rarity_class()
        # then
        self.assertEqual(result, OreRarityClass.R8)

    def test_should_return_R16(self):
        # given
        moon = MoonFactory(create_products=False)
        MoonProductFactory(moon=moon, ore_type=self.ore_type_r4, amount=0.19)
        MoonProductFactory(moon=moon, ore_type=self.ore_type_r16, amount=0.23)
        MoonProductFactory(moon=moon, ore_type=self.ore_type_r8, amount=0.25)
        # when
        result = moon.calc_rarity_class()
        # then
        self.assertEqual(result, OreRarityClass.R16)

    def test_should_return_R32(self):
        # given
        moon = MoonFactory(create_products=False)
        MoonProductFactory(moon=moon, ore_type=self.ore_type_r16, amount=0.23)
        MoonProductFactory(moon=moon, ore_type=self.ore_type_r32, amount=0.19)
        MoonProductFactory(moon=moon, ore_type=self.ore_type_r8, amount=0.25)
        # when
        result = moon.calc_rarity_class()
        # then
        self.assertEqual(result, OreRarityClass.R32)

    def test_should_return_R64(self):
        # given
        moon = MoonFactory(create_products=False)
        MoonProductFactory(moon=moon, ore_type=self.ore_type_r16, amount=0.23)
        MoonProductFactory(moon=moon, ore_type=self.ore_type_r32, amount=0.19)
        MoonProductFactory(moon=moon, ore_type=self.ore_type_r64, amount=0.25)
        # when
        result = moon.calc_rarity_class()
        # then
        self.assertEqual(result, OreRarityClass.R64)

    def test_should_handle_moon_without_products(self):
        # given
        moon = MoonFactory(create_products=False)
        # when
        result = moon.calc_rarity_class()
        # then
        self.assertEqual(result, OreRarityClass.NONE)


class TestMoonProductsSorted(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eveuniverse()

    def setUp(self) -> None:
        helpers.generate_market_prices(use_process_pricing=False)

    def test_should_return_moon_products_in_order(self):
        # given
        moon = MoonFactory()
        # when
        result = moon.products_sorted()
        # then
        ore_types = list(result.values_list("ore_type__name", flat=True))
        self.assertListEqual(["Chromite", "Euxenite", "Xenotime"], ore_types)

    def test_should_handle_products_without_price(self):
        # given
        moon = MoonFactory()
        moon_product = moon.products.first()
        EveMarketPrice.objects.filter(
            eve_type_id=moon_product.ore_type_id
        ).average_price = None
        # when
        result = moon.products_sorted()
        # then
        ore_types = list(result.values_list("ore_type__name", flat=True))
        self.assertListEqual(["Chromite", "Euxenite", "Xenotime"], ore_types)

    def test_should_handle_products_without_amount(self):
        # given
        moon = MoonFactory()
        moon_product = moon.products.first()
        moon_product.amount = 0
        moon_product.save()
        # when
        result = moon.products_sorted()
        # then
        ore_types = list(result.values_list("ore_type__name", flat=True))
        self.assertListEqual(["Chromite", "Euxenite", "Xenotime"], ore_types)

    def test_should_handle_products_without_volume(self):
        # given
        moon = MoonFactory()
        moon_product = moon.products.first()
        volume_backup = moon_product.ore_type.volume
        moon_product.ore_type.volume = None
        moon_product.ore_type.save()
        # when
        result = moon.products_sorted()
        # then
        moon_product.ore_type.volume = volume_backup
        moon_product.ore_type.save()
        ore_types = list(result.values_list("ore_type__name", flat=True))
        self.assertListEqual(["Chromite", "Euxenite", "Xenotime"], ore_types)


class TestMoonOverwriteProducts(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eveuniverse()
        helpers.generate_market_prices()
        load_allianceauth()

    def test_should_overwrite_existing_estimates(self):
        # given
        moon = MoonFactory()
        extraction = CalculatedExtractionFactory()
        ores = {"45506": 7_683_200, "46676": 9_604_000}
        extraction.products = CalculatedExtractionProduct.create_list_from_dict(ores)
        # when
        result = moon.update_products_from_calculated_extraction(extraction)
        # then
        self.assertTrue(result)
        self.assertAlmostEqual(
            moon.products.get(ore_type_id=45506).amount, 0.4, places=2
        )
        self.assertAlmostEqual(
            moon.products.get(ore_type_id=46676).amount, 0.5, places=2
        )
        self.assertIsNone(moon.products_updated_by)
        self.assertIsNotNone(moon.products_updated_at)
        self.assertAlmostEqual(
            moon.products_updated_at, now(), delta=dt.timedelta(minutes=1)
        )

    def test_should_not_overwrite_existing_survey(self):
        # given
        moon = MoonFactory(products_updated_by=UserFactory())
        extraction = CalculatedExtractionFactory()
        ores = {"45506": 7_683_200, "46676": 9_604_000}
        extraction.products = CalculatedExtractionProduct.create_list_from_dict(ores)
        # when
        result = moon.update_products_from_calculated_extraction(extraction)
        # then
        self.assertFalse(result)

    def test_should_overwrite_existing_survey_when_requested(self):
        # given
        moon = MoonFactory(products_updated_by=UserFactory())
        extraction = CalculatedExtractionFactory()
        ores = {"45506": 7_683_200, "46676": 9_604_000}
        extraction.products = CalculatedExtractionProduct.create_list_from_dict(ores)
        # when
        result = moon.update_products_from_calculated_extraction(
            extraction, overwrite_survey=True
        )
        # then
        self.assertTrue(result)

    def test_should_not_overwrite_from_calculated_extraction_without_products(self):
        # given
        moon = MoonFactory()
        extraction = CalculatedExtractionFactory(products=[])
        # when
        result = moon.update_products_from_calculated_extraction(extraction)
        # then
        self.assertFalse(result)
        self.assertTrue(moon.products.exists())

    def test_should_overwrite_products_from_latest_extraction(self):
        # given
        moon = MoonFactory()
        refinery = RefineryFactory(moon=moon)
        ExtractionFactory(refinery=refinery)
        moon.products.all().delete()
        # when
        moon.update_products_from_latest_extraction()
        # then
        self.assertGreater(moon.products.count(), 0)


class TestNotification(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eveuniverse()
        load_allianceauth()

    def test_should_convert_to_calculated_extraction(self):
        # given
        refinery = RefineryFactory()
        my_map = [
            (Extraction.Status.STARTED, CalculatedExtraction.Status.STARTED),
            (Extraction.Status.CANCELED, CalculatedExtraction.Status.CANCELED),
            (Extraction.Status.READY, CalculatedExtraction.Status.READY),
            (Extraction.Status.COMPLETED, CalculatedExtraction.Status.COMPLETED),
        ]
        for in_status, out_status in my_map:
            with self.subTest(status=in_status):
                extraction = ExtractionFactory(status=in_status, refinery=refinery)
                notification = NotificationFactory(extraction=extraction)
                # when
                obj = notification.to_calculated_extraction()
                # then
                self.assertEqual(obj.status, out_status)


class TestOreQualityClass(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eveuniverse()

    def test_should_return_correct_quality(self):
        # given
        ore_quality_regular = EveOreType.objects.get(id=EveTypeId.ZEOLITES)
        ore_quality_improved = EveOreType.objects.get(id=EveTypeId.BRIMFUL_ZEOLITES)
        ore_quality_excellent = EveOreType.objects.get(id=EveTypeId.GLISTENING_ZEOLITES)
        # when/then
        self.assertEqual(ore_quality_regular.quality_class, OreQualityClass.REGULAR)
        self.assertEqual(ore_quality_improved.quality_class, OreQualityClass.IMPROVED)
        self.assertEqual(ore_quality_excellent.quality_class, OreQualityClass.EXCELLENT)

    def test_should_return_correct_tag(self):
        self.assertIn("+100%", OreQualityClass.EXCELLENT.bootstrap_tag_html)


class TestOwner(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eveuniverse()
        load_allianceauth()

    def test_should_return_token(self):
        # given
        owner = OwnerFactory()
        # when
        result = owner.fetch_token()
        # then
        self.assertIsInstance(result, Token)

    def test_should_raise_error_when_no_character_ownership(self):
        # given
        owner = OwnerFactory.build(character_ownership=None)
        # when
        with self.assertRaises(RuntimeError):
            owner.fetch_token()

    def test_should_raise_error_when_no_token_found(self):
        # given
        owner = OwnerFactory()
        Token.objects.filter(user=owner.character_ownership.user).delete()
        # when
        with self.assertRaises(Token.DoesNotExist):
            owner.fetch_token()


@patch(MODELS_PATH + ".esi")
class TestOwnerFetchNotifications(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eveuniverse()
        load_allianceauth()
        helpers.generate_eve_entities_from_allianceauth()

    def test_should_create_new_notifications_from_esi(self, mock_esi):
        # given
        mock_esi.client = esi_client_stub
        _, character_ownership = helpers.create_default_user_from_evecharacter(1005)
        owner = OwnerFactory(character_ownership=character_ownership)
        # when
        owner.fetch_notifications_from_esi()
        # then
        self.assertEqual(owner.notifications.count(), 5)
        obj = owner.notifications.get(notification_id=1005000101)
        self.assertEqual(obj.notif_type, NotificationType.MOONMINING_EXTRACTION_STARTED)
        self.assertEqual(obj.sender_id, 2101)
        self.assertEqual(
            obj.timestamp, dt.datetime(2019, 11, 22, 1, 0, tzinfo=pytz.UTC)
        )
        self.assertEqual(obj.details["moonID"], 40161708)
        self.assertEqual(obj.details["structureID"], 1000000000001)


@patch(MODELS_PATH + ".esi")
@patch(MODELS_PATH + ".notify_admins_throttled", lambda *args, **kwargs: None)
class TestOwnerUpdateRefineries(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eveuniverse()
        load_allianceauth()
        cls.owner = OwnerFactory()

    @patch(
        MODELS_PATH + ".EveSolarSystem.nearest_celestial", new=nearest_celestial_stub
    )
    def test_should_create_new_refineries_from_scratch(self, mock_esi):
        # given
        mock_esi.client = esi_client_stub
        my_eve_moon = EveMoon.objects.get(id=40161708)
        # when
        self.owner.update_refineries_from_esi()
        # then
        self.assertSetEqual(Refinery.objects.ids(), {1000000000001, 1000000000002})
        refinery = Refinery.objects.get(id=1000000000001)
        self.assertEqual(refinery.name, "Auga - Paradise Alpha")
        self.assertEqual(refinery.moon.eve_moon, my_eve_moon)

    @patch(MODELS_PATH + ".EveSolarSystem.nearest_celestial")
    def test_should_handle_OSError_exceptions_from_nearest_celestial(
        self, mock_nearest_celestial, mock_esi
    ):
        # given
        mock_esi.client = esi_client_stub
        mock_nearest_celestial.side_effect = OSError
        # when
        self.owner.update_refineries_from_esi()
        # then
        self.assertSetEqual(Refinery.objects.ids(), {1000000000001, 1000000000002})
        refinery = Refinery.objects.get(id=1000000000001)
        self.assertIsNone(refinery.moon)
        self.assertEqual(mock_nearest_celestial.call_count, 2)

    @patch(
        MODELS_PATH + ".EveSolarSystem.nearest_celestial", new=nearest_celestial_stub
    )
    def test_should_remove_refineries_that_no_longer_exist(self, mock_esi):
        # given
        mock_esi.client = esi_client_stub
        RefineryFactory(id=1990000000001, owner=self.owner)
        # when
        self.owner.update_refineries_from_esi()
        # then
        self.assertSetEqual(Refinery.objects.ids(), {1000000000001, 1000000000002})

    @patch(
        MODELS_PATH + ".EveSolarSystem.nearest_celestial", new=nearest_celestial_stub
    )
    def test_should_not_remove_refineries_after_OSError_in_corporation_structures(
        self, mock_esi
    ):
        # given
        mock_esi.client.Corporation.get_corporations_corporation_id_structures.side_effect = (
            OSError
        )
        RefineryFactory(id=1990000000001, owner=self.owner)
        # when
        with self.assertRaises(OSError):
            self.owner.update_refineries_from_esi()
        # then
        self.assertSetEqual(Refinery.objects.ids(), {1990000000001})

    @patch(
        MODELS_PATH + ".EveSolarSystem.nearest_celestial", new=nearest_celestial_stub
    )
    def test_should_continue_with_other_refineries_after_OS_error(self, mock_esi):
        def my_get_corporations_corporation_id_structures(*args, **kwargs):
            """Pass through"""
            return (
                esi_client_stub.Corporation.get_corporations_corporation_id_structures(
                    *args, **kwargs
                )
            )

        def my_get_universe_structures_structure_id(*args, **kwargs):
            """Return OS error for specific structure only."""
            if "structure_id" in kwargs and kwargs["structure_id"] == 1000000000001:
                raise OSError
            return esi_client_stub.Universe.get_universe_structures_structure_id(
                *args, **kwargs
            )

        # given
        mock_esi.client.Corporation.get_corporations_corporation_id_structures = (
            my_get_corporations_corporation_id_structures
        )
        mock_esi.client.Universe.get_universe_structures_structure_id = (
            my_get_universe_structures_structure_id
        )
        RefineryFactory(id=1000000000001, owner=self.owner)
        RefineryFactory(id=1000000000002, owner=self.owner)
        # when
        self.owner.update_refineries_from_esi()
        # then
        self.assertSetEqual(Refinery.objects.ids(), {1000000000001, 1000000000002})


@patch(MODELS_PATH + ".esi")
class TestOwnerUpdateExtractions(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eveuniverse()
        load_allianceauth()
        helpers.generate_eve_entities_from_allianceauth()

    def test_should_create_started_extraction_with_products(self, mock_esi):
        # given
        mock_esi.client = esi_client_stub
        owner = OwnerFactory()
        owner.fetch_notifications_from_esi()
        refinery = RefineryFactory(id=1000000000001, owner=owner)
        # when
        owner.update_extractions()
        # then
        self.assertEqual(refinery.extractions.count(), 1)
        extraction = refinery.extractions.first()
        self.assertEqual(extraction.status, Extraction.Status.STARTED)
        self.assertEqual(
            extraction.chunk_arrival_at,
            dt.datetime(2021, 4, 15, 18, 0, tzinfo=pytz.UTC),
        )
        self.assertEqual(extraction.started_by_id, 1001)
        self.assertEqual(extraction.products.count(), 4)
        product = extraction.products.get(ore_type_id=45506)
        self.assertEqual(product.volume, 1288475.124715103)
        product = extraction.products.get(ore_type_id=46676)
        self.assertEqual(product.volume, 544691.7637724016)
        product = extraction.products.get(ore_type_id=22)
        self.assertEqual(product.volume, 526825.4047522942)
        product = extraction.products.get(ore_type_id=46689)
        self.assertEqual(product.volume, 528996.6386983792)
        self.assertIsNotNone(extraction.value)
        self.assertIsNotNone(extraction.is_jackpot)


@patch(MODELS_PATH + ".esi")
class TestOwnerUpdateExtractionsFromEsi(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eveuniverse()
        load_allianceauth()
        helpers.generate_eve_entities_from_allianceauth()
        cls.owner = OwnerFactory()

    def test_should_create_started_extraction(self, mock_esi):
        # given
        mock_esi.client = esi_client_stub
        refinery = RefineryFactory(id=1000000000001, owner=self.owner)
        # when
        with patch(MODELS_PATH + ".now") as mock_now:
            mock_now.return_value = dt.datetime(2021, 4, 5, 12, 0, 0, tzinfo=pytz.UTC)
            self.owner.update_extractions_from_esi()
        # then
        self.assertEqual(refinery.extractions.count(), 1)
        extraction = refinery.extractions.first()
        self.assertEqual(extraction.status, Extraction.Status.STARTED)
        self.assertEqual(
            extraction.chunk_arrival_at,
            dt.datetime(2021, 4, 15, 18, 0, tzinfo=pytz.UTC),
        )
        self.assertEqual(
            extraction.started_at, dt.datetime(2021, 4, 1, 12, 00, tzinfo=pytz.UTC)
        )
        self.assertEqual(
            extraction.auto_fracture_at,
            dt.datetime(2021, 4, 15, 21, 00, tzinfo=pytz.UTC),
        )
        self.assertEqual(extraction.products.count(), 0)
        self.assertIsNone(extraction.value)
        self.assertIsNone(extraction.is_jackpot)

    def test_should_create_completed_extraction(self, mock_esi):
        # given
        mock_esi.client = esi_client_stub
        refinery = RefineryFactory(id=1000000000001, owner=self.owner)
        # when
        with patch(MODELS_PATH + ".now") as mock_now:
            mock_now.return_value = dt.datetime(2021, 4, 18, 18, 15, 0, tzinfo=pytz.UTC)
            self.owner.update_extractions_from_esi()
        # then
        self.assertEqual(refinery.extractions.count(), 1)
        extraction = refinery.extractions.first()
        self.assertEqual(extraction.status, Extraction.Status.COMPLETED)

    def test_should_identify_canceled_extractions(self, mock_esi):
        # given
        mock_esi.client = esi_client_stub
        refinery = RefineryFactory(id=1000000000001, owner=self.owner)
        started_extraction = ExtractionFactory(
            refinery=refinery,
            started_at=dt.datetime(2021, 3, 10, 18, 0, tzinfo=pytz.UTC),
            chunk_arrival_at=dt.datetime(2021, 3, 15, 18, 0, tzinfo=pytz.UTC),
            auto_fracture_at=dt.datetime(2021, 3, 15, 21, 0, tzinfo=pytz.UTC),
            status=Extraction.Status.STARTED,
            create_products=False,
        )
        # when
        with patch(MODELS_PATH + ".now") as mock_now:
            mock_now.return_value = dt.datetime(2021, 4, 1, 12, 0, tzinfo=pytz.UTC)
            self.owner.update_extractions_from_esi()
        # then
        started_extraction.refresh_from_db()
        self.assertEqual(started_extraction.status, Extraction.Status.CANCELED)
        self.assertTrue(started_extraction.canceled_at)


@patch(MODELS_PATH + ".esi")
class TestOwnerUpdateExtractionsFromNotifications(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eveuniverse()
        load_allianceauth()
        helpers.generate_eve_entities_from_allianceauth()

    def test_should_update_started_extraction_with_products(self, mock_esi):
        # given
        mock_esi.client = esi_client_stub
        owner = OwnerFactory()
        refinery = RefineryFactory(owner=owner)
        extraction = ExtractionFactory(
            refinery=refinery, create_products=False, status=Extraction.Status.STARTED
        )
        calc_extraction = extraction.to_calculated_extraction()
        NotificationFactory2(
            extraction=calc_extraction, owner=owner, create_products=True
        )
        # when
        owner.update_extractions_from_notifications()
        # then
        extraction.refresh_from_db()
        self.assertEqual(extraction.status, Extraction.Status.STARTED)
        self.assertEqual(
            set(extraction.products.values_list("ore_type_id", flat=True)),
            {EveTypeId.CHROMITE, EveTypeId.EUXENITE, EveTypeId.XENOTIME},
        )

    def test_should_create_canceled_extraction_with_products(self, mock_esi):
        # given
        mock_esi.client = esi_client_stub
        owner = OwnerFactory()
        owner.fetch_notifications_from_esi()
        refinery = RefineryFactory(id=1000000000002, owner=owner)
        extraction = ExtractionFactory(
            refinery=refinery,
            chunk_arrival_at=dt.datetime(2021, 4, 15, 18, 0, tzinfo=pytz.UTC),
            started_at=dt.datetime(2021, 4, 10, 18, 0, tzinfo=pytz.UTC),
            auto_fracture_at=dt.datetime(2021, 4, 15, 21, 0, tzinfo=pytz.UTC),
            create_products=False,
        )
        # when
        owner.update_extractions_from_notifications()
        # then
        self.assertEqual(refinery.extractions.count(), 1)
        extraction.refresh_from_db()
        self.assertEqual(extraction.status, Extraction.Status.CANCELED)
        self.assertEqual(
            extraction.canceled_at,
            dt.datetime(2019, 11, 22, 2, tzinfo=pytz.UTC),
        )
        self.assertEqual(extraction.canceled_by_id, 1001)
        self.assertEqual(
            set(extraction.products.values_list("ore_type_id", flat=True)),
            {45506, 46676, 22, 46689},
        )

    def test_should_create_finished_extraction_with_products(self, mock_esi):
        # given
        mock_esi.client = esi_client_stub
        owner = OwnerFactory()
        owner.fetch_notifications_from_esi()
        refinery = RefineryFactory(id=1000000000003, owner=owner)
        extraction = ExtractionFactory(
            refinery=refinery,
            chunk_arrival_at=dt.datetime(2021, 4, 15, 18, 0, tzinfo=pytz.UTC),
            started_at=dt.datetime(2021, 4, 10, 18, 0, tzinfo=pytz.UTC),
            auto_fracture_at=dt.datetime(2021, 4, 15, 21, 0, tzinfo=pytz.UTC),
            create_products=False,
        )
        ExtractionProductFactory(
            extraction=extraction, ore_type_id=45506, volume=1288475
        )
        ExtractionProductFactory(
            extraction=extraction, ore_type_id=46676, volume=544691
        )
        ExtractionProductFactory(extraction=extraction, ore_type_id=22, volume=526825)
        ExtractionProductFactory(
            extraction=extraction, ore_type_id=46689, volume=528996
        )
        # when
        owner.update_extractions_from_notifications()
        # then
        self.assertEqual(refinery.extractions.count(), 1)
        extraction.refresh_from_db()
        self.assertEqual(extraction.status, Extraction.Status.READY)
        self.assertEqual(
            set(extraction.products.values_list("ore_type_id", flat=True)),
            {46311, 46676, 46678, 46689},
        )

    def test_should_create_manually_fractured_extraction_with_products(self, mock_esi):
        # given
        mock_esi.client = esi_client_stub
        owner = OwnerFactory()
        owner.fetch_notifications_from_esi()
        refinery = RefineryFactory(id=1000000000004, owner=owner)
        extraction = ExtractionFactory(
            refinery=refinery,
            chunk_arrival_at=dt.datetime(2021, 4, 15, 18, 0, tzinfo=pytz.UTC),
            started_at=dt.datetime(2021, 4, 10, 18, 0, tzinfo=pytz.UTC),
            auto_fracture_at=dt.datetime(2021, 4, 15, 21, 0, tzinfo=pytz.UTC),
            create_products=False,
        )
        ExtractionProductFactory(
            extraction=extraction, ore_type_id=45506, volume=1288475
        )
        ExtractionProductFactory(
            extraction=extraction, ore_type_id=46676, volume=544691
        )
        ExtractionProductFactory(extraction=extraction, ore_type_id=22, volume=526825)
        ExtractionProductFactory(
            extraction=extraction, ore_type_id=46689, volume=528996
        )
        # when
        owner.update_extractions_from_notifications()
        # then
        self.assertEqual(refinery.extractions.count(), 1)
        extraction.refresh_from_db()
        self.assertEqual(extraction.status, Extraction.Status.COMPLETED)
        self.assertEqual(extraction.fractured_by_id, 1001)
        self.assertEqual(
            set(extraction.products.values_list("ore_type_id", flat=True)),
            {46311, 46676, 46678, 46689},
        )

    def test_should_create_auto_fractured_extraction_with_products(self, mock_esi):
        # given
        mock_esi.client = esi_client_stub
        owner = OwnerFactory()
        owner.fetch_notifications_from_esi()
        refinery = RefineryFactory(id=1000000000005, owner=owner)
        extraction = ExtractionFactory(
            refinery=refinery,
            chunk_arrival_at=dt.datetime(2021, 4, 15, 18, 0, tzinfo=pytz.UTC),
            started_at=dt.datetime(2021, 4, 10, 18, 0, tzinfo=pytz.UTC),
            auto_fracture_at=dt.datetime(2021, 4, 15, 21, 0, tzinfo=pytz.UTC),
            create_products=False,
        )
        ExtractionProductFactory(
            extraction=extraction, ore_type_id=45506, volume=1288475
        )
        ExtractionProductFactory(
            extraction=extraction, ore_type_id=46676, volume=544691
        )
        ExtractionProductFactory(extraction=extraction, ore_type_id=22, volume=526825)
        ExtractionProductFactory(
            extraction=extraction, ore_type_id=46689, volume=528996
        )
        # when
        owner.update_extractions_from_notifications()
        # then
        self.assertEqual(refinery.extractions.count(), 1)
        extraction.refresh_from_db()
        self.assertEqual(extraction.status, Extraction.Status.COMPLETED)
        self.assertIsNone(extraction.fractured_by)
        self.assertEqual(
            set(extraction.products.values_list("ore_type_id", flat=True)),
            {46311, 46676, 46678},
        )

    def test_should_create_manually_fractured_extraction_with_products_2(
        self, mock_esi
    ):
        """notification chain starting with 'finished' and missing 'started'"""
        # given
        mock_esi.client = esi_client_stub
        owner = OwnerFactory()
        owner.fetch_notifications_from_esi()
        refinery = RefineryFactory(id=1000000000006, owner=owner)
        extraction = ExtractionFactory(
            refinery=refinery,
            chunk_arrival_at=dt.datetime(2021, 4, 15, 18, 0, tzinfo=pytz.UTC),
            started_at=dt.datetime(2021, 4, 10, 18, 0, tzinfo=pytz.UTC),
            auto_fracture_at=dt.datetime(2021, 4, 15, 21, 00, tzinfo=pytz.UTC),
            status=Extraction.Status.STARTED,
            create_products=False,
        )
        # when
        owner.update_extractions_from_notifications()
        # then
        self.assertEqual(refinery.extractions.count(), 1)
        extraction.refresh_from_db()
        self.assertEqual(extraction.status, Extraction.Status.COMPLETED)
        self.assertEqual(extraction.fractured_by_id, 1001)
        self.assertEqual(
            set(extraction.products.values_list("ore_type_id", flat=True)),
            {46311, 46676, 46678, 46689},
        )

    def test_should_cancel_existing_extraction(self, mock_esi):
        # given
        mock_esi.client = esi_client_stub
        _, character_ownership = helpers.create_default_user_from_evecharacter(1002)
        owner = OwnerFactory(character_ownership=character_ownership)
        owner.fetch_notifications_from_esi()
        refinery = RefineryFactory(id=1000000000001, owner=owner)
        extraction = ExtractionFactory(
            refinery=refinery,
            chunk_arrival_at=dt.datetime(2021, 4, 15, 18, 0, tzinfo=pytz.UTC),
            started_at=dt.datetime(2021, 4, 10, 18, 0, tzinfo=pytz.UTC),
            auto_fracture_at=dt.datetime(2021, 4, 15, 21, 0, tzinfo=pytz.UTC),
            create_products=False,
        )
        # when
        owner.update_extractions_from_notifications()
        # then
        self.assertEqual(refinery.extractions.count(), 1)
        extraction.refresh_from_db()
        self.assertEqual(extraction.status, Extraction.Status.CANCELED)

    def test_should_cancel_existing_and_update_two_other(self, mock_esi):
        # given
        mock_esi.client = esi_client_stub
        _, character_ownership = helpers.create_default_user_from_evecharacter(1004)
        owner = OwnerFactory(character_ownership=character_ownership)
        owner.fetch_notifications_from_esi()
        refinery = RefineryFactory(id=1000000000001, owner=owner)
        ready_time_1 = dt.datetime(2019, 11, 21, 10, tzinfo=pytz.UTC)
        ready_time_2 = dt.datetime(2019, 11, 21, 11, tzinfo=pytz.UTC)
        ready_time_3 = dt.datetime(2019, 11, 21, 12, tzinfo=pytz.UTC)
        extraction_1 = ExtractionFactory(
            refinery=refinery,
            chunk_arrival_at=ready_time_1,
            started_at=ready_time_1 - dt.timedelta(days=14),
            auto_fracture_at=ready_time_1 + dt.timedelta(hours=4),
            status=Extraction.Status.STARTED,
            create_products=False,
        )
        extraction_2 = ExtractionFactory(
            refinery=refinery,
            chunk_arrival_at=ready_time_2,
            started_at=ready_time_2 - dt.timedelta(days=14),
            auto_fracture_at=ready_time_2 + dt.timedelta(hours=4),
            status=Extraction.Status.STARTED,
            create_products=False,
        )
        ExtractionProductFactory(
            extraction=extraction_2, ore_type_id=45506, volume=1288475
        )
        ExtractionProductFactory(
            extraction=extraction_2, ore_type_id=46676, volume=544691
        )
        ExtractionProductFactory(extraction=extraction_2, ore_type_id=22, volume=526825)
        ExtractionProductFactory(
            extraction=extraction_2, ore_type_id=46689, volume=528996
        )
        extraction_3 = ExtractionFactory(
            refinery=refinery,
            chunk_arrival_at=ready_time_3,
            started_at=ready_time_3 - dt.timedelta(days=14),
            auto_fracture_at=ready_time_3 + dt.timedelta(hours=4),
            status=Extraction.Status.STARTED,
            create_products=False,
        )
        # when
        owner.update_extractions_from_notifications()
        # then
        self.assertEqual(refinery.extractions.count(), 3)
        extraction_1.refresh_from_db()
        self.assertEqual(extraction_1.status, Extraction.Status.CANCELED)
        extraction_2.refresh_from_db()
        self.assertEqual(extraction_2.status, Extraction.Status.COMPLETED)
        self.assertEqual(
            set(extraction_2.products.values_list("ore_type_id", flat=True)),
            {46311, 46676, 46678, 46689},
        )
        extraction_3.refresh_from_db()
        self.assertEqual(extraction_3.status, Extraction.Status.STARTED)
        self.assertEqual(
            set(extraction_3.products.values_list("ore_type_id", flat=True)),
            {45506, 46676, 22, 46689},
        )

    def test_should_update_refinery_with_moon_from_notification_if_not_found(
        self, mock_esi
    ):
        # given
        mock_esi.client = esi_client_stub
        owner = OwnerFactory()
        owner.fetch_notifications_from_esi()
        refinery = RefineryFactory(id=1000000000001, moon=None, owner=owner)
        # when
        owner.update_extractions()
        # then
        refinery.refresh_from_db()
        self.assertEqual(refinery.moon.pk, 40161708)

    def test_should_update_moon_products_when_no_survey_exists(self, mock_esi):
        # given
        mock_esi.client = esi_client_stub
        moon = MoonFactory()
        moon.products.first().delete()
        owner = OwnerFactory()
        refinery = RefineryFactory(owner=owner, moon=moon)
        extraction = ExtractionFactory(
            refinery=refinery, status=Extraction.Status.STARTED
        )
        calc_extraction = extraction.to_calculated_extraction()
        NotificationFactory2(
            extraction=calc_extraction, owner=owner, create_products=True
        )
        # when
        owner.update_extractions_from_notifications()
        # then
        self.assertEqual(moon.products.count(), 3)

    @patch(MODELS_PATH + ".MOONMINING_OVERWRITE_SURVEYS_WITH_ESTIMATES", False)
    def test_should_not_update_moon_products_when_survey_exists(self, mock_esi):
        # given
        mock_esi.client = esi_client_stub
        moon = MoonFactory(products_updated_by=UserFactory())
        moon.products.first().delete()
        owner = OwnerFactory()
        refinery = RefineryFactory(owner=owner, moon=moon)
        extraction = ExtractionFactory(
            refinery=refinery, status=Extraction.Status.STARTED
        )
        calc_extraction = extraction.to_calculated_extraction()
        NotificationFactory2(
            extraction=calc_extraction, owner=owner, create_products=True
        )
        # when
        owner.update_extractions_from_notifications()
        # then
        self.assertEqual(moon.products.count(), 2)

    @patch(MODELS_PATH + ".MOONMINING_OVERWRITE_SURVEYS_WITH_ESTIMATES", True)
    def test_should_update_moon_products_when_survey_exists_alternate(self, mock_esi):
        # given
        mock_esi.client = esi_client_stub
        moon = MoonFactory(products_updated_by=UserFactory())
        moon.products.first().delete()
        owner = OwnerFactory()
        refinery = RefineryFactory(owner=owner, moon=moon)
        extraction = ExtractionFactory(
            refinery=refinery, status=Extraction.Status.STARTED
        )
        calc_extraction = extraction.to_calculated_extraction()
        NotificationFactory2(
            extraction=calc_extraction, owner=owner, create_products=True
        )
        # when
        owner.update_extractions_from_notifications()
        # then
        self.assertEqual(moon.products.count(), 3)


@patch(MODELS_PATH + ".esi")
class TestOwnerUpdateMiningLedger(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eveuniverse()
        load_allianceauth()
        helpers.generate_eve_entities_from_allianceauth()

    def test_should_return_observer_ids_from_esi(self, mock_esi):
        # given
        mock_esi.client = esi_client_stub
        owner = OwnerFactory()
        # when
        result = owner.fetch_mining_ledger_observers_from_esi()
        # then
        self.assertSetEqual(result, {1000000000001, 1000000000002})

    def test_should_create_new_mining_ledger(self, mock_esi):
        # given
        mock_esi.client = esi_client_stub
        owner = OwnerFactory()
        refinery = RefineryFactory(id=1000000000001, owner=owner)
        # when
        refinery.update_mining_ledger_from_esi()
        # then
        refinery.refresh_from_db()
        self.assertTrue(refinery.ledger_last_update_ok)
        self.assertAlmostEqual(
            refinery.ledger_last_update_at, now(), delta=dt.timedelta(minutes=1)
        )
        self.assertEqual(refinery.mining_ledger.count(), 2)
        obj = refinery.mining_ledger.get(character_id=1001)
        self.assertEqual(obj.day, dt.date(2017, 9, 19))
        self.assertEqual(obj.quantity, 500)
        self.assertEqual(obj.corporation_id, 2001)
        self.assertEqual(obj.ore_type_id, 45506)

    def test_should_update_existing_mining_ledger(self, mock_esi):
        # given
        mock_esi.client = esi_client_stub
        owner = OwnerFactory()
        refinery = RefineryFactory(id=1000000000001, owner=owner)
        MiningLedgerRecordFactory(
            refinery=refinery,
            day=dt.date(2017, 9, 19),
            character_id=1001,
            ore_type_id=EveTypeId.CINNABAR,
            quantity=199,
        )
        # when
        refinery.update_mining_ledger_from_esi()
        # then
        self.assertEqual(refinery.mining_ledger.count(), 2)
        obj = refinery.mining_ledger.get(character_id=1001)
        self.assertEqual(obj.quantity, 500)

    def test_should_mark_when_update_failed(self, mock_esi):
        # given
        mock_esi.client.Industry.get_corporation_corporation_id_mining_observers_observer_id.side_effect = (
            OSError
        )
        owner = OwnerFactory()
        refinery = RefineryFactory(id=1000000000001, owner=owner)
        # when
        with self.assertRaises(OSError):
            refinery.update_mining_ledger_from_esi()
        # then
        refinery.refresh_from_db()
        self.assertFalse(refinery.ledger_last_update_ok)
        self.assertAlmostEqual(
            refinery.ledger_last_update_at, now(), delta=dt.timedelta(minutes=1)
        )



from celery import chain, shared_task

from django.contrib.auth.models import User
from django.utils.timezone import now
from eveuniverse.models import EveMarketPrice

from allianceauth.services.hooks import get_extension_logger
from app_utils.esi import fetch_esi_status
from app_utils.logging import LoggerAddTag

from . import __title__
from .models import EveOreType, Extraction, Moon, Owner, Refinery

logger = LoggerAddTag(get_extension_logger(__name__), __title__)

TASK_PRIORITY_LOWER = 6


@shared_task
def process_survey_input(scans, user_pk=None) -> bool:
    """Update moons from survey input."""
    user = User.objects.get(pk=user_pk) if user_pk else None
    return Moon.objects.update_moons_from_survey(scans, user)


@shared_task
def run_regular_updates():
    """Run main tasks for regular updates."""
    owners_to_update = Owner.objects.filter(is_enabled=True)
    owner_pks = owners_to_update.values_list("pk", flat=True)
    logger.info("Updating %d owners...", len(owner_pks))
    owners_to_update.update(last_update_ok=None, last_update_at=now())
    for owner_pk in owner_pks:
        update_owner.delay(owner_pk)


@shared_task
def update_owner(owner_pk):
    """Update refineries and extractions for given owner."""
    if fetch_esi_status().is_ok:
        chain(
            update_refineries_from_esi_for_owner.si(owner_pk),
            fetch_notifications_from_esi_for_owner.si(owner_pk),
            update_extractions_for_owner.si(owner_pk),
            mark_successful_update_for_owner.si(owner_pk),
        ).delay()
    else:
        logger.warning("ESI ist not available. Aborting.")


@shared_task
def update_refineries_from_esi_for_owner(owner_pk):
    """Update refineries for a owner from ESI."""
    owner = Owner.objects.get(pk=owner_pk)
    owner.update_refineries_from_esi()


@shared_task
def fetch_notifications_from_esi_for_owner(owner_pk):
    """Update extractions for a owner from ESI."""
    owner = Owner.objects.get(pk=owner_pk)
    owner.fetch_notifications_from_esi()


@shared_task
def update_extractions_for_owner(owner_pk):
    """Update extractions for a owner from ESI."""
    owner = Owner.objects.get(pk=owner_pk)
    owner.update_extractions()


@shared_task
def mark_successful_update_for_owner(owner_pk):
    """Mark a successful update for this corporation."""
    owner = Owner.objects.get(pk=owner_pk)
    owner.last_update_ok = True
    owner.save()


@shared_task
def run_report_updates():
    """Run tasks for updating reports and related data."""
    owners_to_update = Owner.objects.filter(is_enabled=True)
    owner_pks = owners_to_update.values_list("pk", flat=True)
    logger.info("Updating mining ledgers for %d owners...", len(owner_pks))
    for owner_pk in owner_pks:
        update_mining_ledger_for_owner.delay(owner_pk)


@shared_task
def update_mining_ledger_for_owner(owner_pk):
    """Update mining ledger for a owner from ESI."""
    if fetch_esi_status().is_ok:
        owner = Owner.objects.get(pk=owner_pk)
        observer_ids = owner.fetch_mining_ledger_observers_from_esi()
        for refinery_id in owner.refineries.filter(id__in=observer_ids).values_list(
            "id", flat=True
        ):
            update_mining_ledger_for_refinery.apply_async(
                kwargs={"refinery_id": refinery_id}, priority=TASK_PRIORITY_LOWER
            )
    else:
        logger.warning("ESI ist not available. Aborting.")


@shared_task
def update_mining_ledger_for_refinery(refinery_id):
    """Update mining ledger for a refinery from ESI."""
    refinery = Refinery.objects.get(id=refinery_id)
    refinery.update_mining_ledger_from_esi()


@shared_task
def run_calculated_properties_update():
    """Update the calculated properties of all moons and all extractions."""
    if fetch_esi_status().is_ok:
        chain(
            update_market_prices.si().set(priority=TASK_PRIORITY_LOWER),
            update_current_ore_prices.si().set(priority=TASK_PRIORITY_LOWER),
            update_moons.si().set(priority=TASK_PRIORITY_LOWER),
            update_extractions.si().set(priority=TASK_PRIORITY_LOWER),
        ).delay()
    else:
        logger.warning("ESI ist not available. Aborting.")


@shared_task
def update_market_prices():
    """Update all market prices."""
    EveMarketPrice.objects.update_from_esi()


@shared_task
def update_current_ore_prices():
    """Update current prices for all ore types."""
    EveOreType.objects.update_current_prices()


@shared_task
def update_moons():
    """Update the calculated properties of all moons."""
    moon_pks = Moon.objects.values_list("pk", flat=True)
    logger.info("Updating calculated properties for %d moons ...", len(moon_pks))
    for moon_pk in moon_pks:
        update_moon_calculated_properties.apply_async(
            kwargs={"moon_pk": moon_pk}, priority=TASK_PRIORITY_LOWER
        )


@shared_task
def update_moon_calculated_properties(moon_pk):
    """Update all calculated properties for given moon."""
    moon = Moon.objects.get(pk=moon_pk)
    moon.update_calculated_properties()


@shared_task
def update_moon_products_from_latest_extraction(moon_pk):
    """Update moon products from latest extraction."""
    moon = Moon.objects.get(pk=moon_pk)
    result = moon.update_products_from_latest_extraction(overwrite_survey=True)
    if result is True:
        logger.info("%s: Updated moon products from latest extraction", moon)
    elif result is None:
        logger.info(
            "%s: Failed to update moon products from latest extraction, "
            "because this moon has no extractions.",
            moon,
        )
    else:
        logger.info("%s: Failed to update moon products from latest extraction", moon)


@shared_task
def update_extractions():
    """Update the calculated properties of all extractions."""
    extraction_pks = Extraction.objects.values_list("pk", flat=True)
    logger.info(
        "Updating calculated properties for %d extractions ...", len(extraction_pks)
    )
    for extraction_pk in extraction_pks:
        update_extraction_calculated_properties.apply_async(
            kwargs={"extraction_pk": extraction_pk}, priority=TASK_PRIORITY_LOWER
        )


@shared_task
def update_extraction_calculated_properties(extraction_pk):
    """Update all calculated properties for given extraction."""
    extraction = Extraction.objects.get(pk=extraction_pk)
    extraction.update_calculated_properties()



from app_utils.django import clean_setting

# Number of hours an extractions that has passed its ready time is still shown
# on the upcoming extractions tab
MOONMINING_COMPLETED_EXTRACTIONS_HOURS_UNTIL_STALE = clean_setting(
    "MOONMINING_COMPLETED_EXTRACTIONS_HOURS_UNTIL_STALE", 12
)

# Reprocessing yield used for calculating reprocess prices
MOONMINING_REPROCESSING_YIELD = clean_setting("MOONMINING_REPROCESSING_YIELD", 0.85)

# Whether to calculate prices from reprocessed materials or not.
# Will use direct ore prices when switched off
MOONMINING_USE_REPROCESS_PRICING = clean_setting(
    "MOONMINING_USE_REPROCESS_PRICING", False
)

# Total ore volume per month used for calculating moon values.
MOONMINING_VOLUME_PER_DAY = clean_setting("MOONMINING_VOLUME_PER_DAY", 960_400)
MOONMINING_DAYS_PER_MONTH = clean_setting("MOONMINING_DAYS_PER_MONTH", 30.4)
MOONMINING_VOLUME_PER_MONTH = MOONMINING_VOLUME_PER_DAY * MOONMINING_DAYS_PER_MONTH

# whether admins will get notifications about important events like
# when someone adds a new owner
MOONMINING_ADMIN_NOTIFICATIONS_ENABLED = clean_setting(
    "MOONMINING_ADMIN_NOTIFICATIONS_ENABLED", True
)

# whether uploaded survey are automatically overwritten by product estimates from
# extractions to keep the moon values current
MOONMINING_OVERWRITE_SURVEYS_WITH_ESTIMATES = clean_setting(
    "MOONMINING_OVERWRITE_SURVEYS_WITH_ESTIMATES", False
)



from django.test import TestCase
from eveuniverse.models import EveType
from eveuniverse.tools.testdata import ModelSpec, create_testdata

from moonmining.constants import EveCategoryId, EveGroupId

from . import test_data_filename


class CreateEveUniverseTestData(TestCase):
    def test_create_testdata(self):
        testdata_spec = [
            ModelSpec(
                "EveGroup",
                ids=[EveGroupId.MOON.value, EveGroupId.MINERAL.value],
                include_children=True,
            ),
            ModelSpec(
                "EveCategory",
                ids=[EveCategoryId.ASTEROID.value],
                include_children=True,
                enabled_sections=[
                    EveType.Section.TYPE_MATERIALS,
                    EveType.Section.DOGMAS,
                ],
            ),
            ModelSpec(
                "EveType",
                ids=[
                    35834,  # Keepstar
                    35835,  # Athanor,
                ],
            ),
            ModelSpec(
                "EveMoon",
                ids=[
                    40161708,  # Auga V - Moon 1
                    40161709,  # Auga V - Moon 2
                    40131695,  # Helgatild IX - Moon 12
                ],
            ),
        ]
        create_testdata(testdata_spec, test_data_filename())



import datetime as dt
from dataclasses import dataclass
from enum import IntEnum, auto
from typing import List

from . import helpers


@dataclass
class CalculatedExtraction:
    """An extraction calculated from moon mining notifications."""

    class Status(IntEnum):
        STARTED = auto()
        CANCELED = auto()
        READY = auto()
        COMPLETED = auto()
        UNDEFINED = auto()

    refinery_id: int
    status: Status
    auto_fracture_at: dt.datetime = None
    canceled_at: dt.datetime = None
    canceled_by: int = None
    chunk_arrival_at: dt.datetime = None
    fractured_at: dt.datetime = None
    fractured_by: int = None
    products: List["CalculatedExtractionProduct"] = None
    started_at: dt.datetime = None
    started_by: int = None

    def __post_init__(self):
        self.refinery_id = int(self.refinery_id)
        self.status = self.Status(self.status)
        if self.started_at:
            self.started_at = helpers.round_seconds(self.started_at)
        if self.chunk_arrival_at:
            self.chunk_arrival_at = helpers.round_seconds(self.chunk_arrival_at)
        if self.auto_fracture_at:
            self.auto_fracture_at = helpers.round_seconds(self.auto_fracture_at)

    @property
    def duration(self) -> dt.timedelta:
        if self.chunk_arrival_at and self.started_at:
            return self.chunk_arrival_at - self.started_at
        raise ValueError("chunk_arrival_at and/or started_at not defined")

    def total_volume(self) -> float:
        if not self.products:
            return 0
        total = 0.0
        for product in self.products:
            total += product.volume
        return total

    def moon_products_estimated(
        self, volume_per_day: float
    ) -> List["CalculatedMoonProduct"]:
        """List of products with estimated amounts."""
        duration_in_days = self.duration.total_seconds() / (60 * 60 * 24)
        if duration_in_days <= 0:
            raise ValueError("Can not estimate products without duration.")
        max_volume = duration_in_days * volume_per_day
        correction_factor = max(1, self.total_volume() / max_volume)
        products = [
            CalculatedMoonProduct(
                ore_type_id=product.ore_type_id,
                amount=product.calculated_share(duration_in_days, volume_per_day)
                / correction_factor,
            )
            for product in self.products
        ]
        return products


@dataclass
class CalculatedExtractionProduct:
    """Product of an extraction calculated from moon mining notifications."""

    ore_type_id: int
    volume: float

    def __post_init__(self):
        self.ore_type_id = int(self.ore_type_id)

    def calculated_share(self, duration_in_days: float, volume_per_day: float) -> float:
        """Calculate share of these moon products"""
        return self.volume / (duration_in_days * volume_per_day)

    @classmethod
    def create_list_from_dict(cls, ores: dict) -> List["CalculatedExtractionProduct"]:
        return [cls(ore_type_id, volume) for ore_type_id, volume in ores.items()]


@dataclass
class CalculatedMoonProduct:
    """Product of an extraction calculated from moon mining notifications."""

    ore_type_id: int
    amount: float

    def __post_init__(self):
        self.ore_type_id = int(self.ore_type_id)



import csv
from pathlib import Path

from bravado.exception import HTTPBadGateway, HTTPGatewayTimeout, HTTPServiceUnavailable

from django.core.management.base import BaseCommand, CommandError
from django.db import IntegrityError, transaction
from eveuniverse.core.esitools import is_esi_online
from eveuniverse.models import EveMoon

from allianceauth.services.hooks import get_extension_logger
from app_utils.logging import LoggerAddTag

from moonmining.models import EveOreType, Moon, MoonProduct

from ... import __title__, tasks

MAX_RETRIES = 3
BULK_BATCH_SIZE = 500


logger = LoggerAddTag(get_extension_logger(__name__), __title__)


class Command(BaseCommand):
    help = "Import moons from an CSV file."

    def add_arguments(self, parser):
        parser.add_argument("input_file", help="Path of CSV file to be imported")
        parser.add_argument(
            "--force-refetch",
            action="store_const",
            const=True,
            default=False,
            help="When set all needed eveuniverse objects will be fetched again from ESI",
        )
        parser.add_argument(
            "--force-update",
            action="store_const",
            const=True,
            default=False,
            help="When set all Moon and MoonProduct objects will be updated from the input file",
        )
        parser.add_argument(
            "--force-calc",
            action="store_const",
            const=True,
            default=False,
            help="When set will always re-calculate income",
        )
        parser.add_argument(
            "--disable-esi-check",
            action="store_const",
            const=False,
            default=False,
            help="When set script will not check if ESI is online",
        )

    def handle(self, *args, **options):
        if not options["disable_esi_check"] and not is_esi_online():
            raise CommandError("ESI if offline. Aborting")

        input_file = Path(options["input_file"])
        if not input_file.exists():
            raise CommandError(
                f"Could not find a file with the path: {input_file.resolve()}"
            )
        moons, ore_types = self.read_moons(input_file)
        self.fetch_missing_eve_objects(
            EveModel=EveMoon,
            ids_incoming=set(moons.keys()),
            force_refetch=options["force_refetch"],
        )
        self.fetch_missing_eve_objects(
            EveModel=EveOreType,
            ids_incoming=ore_types,
            force_refetch=options["force_refetch"],
        )
        self.import_moons(moons, options["force_update"])
        self.update_moons(moons)
        self.stdout.write(
            self.style.SUCCESS(
                "Import completed. Recalculation of imported moons started."
            )
        )

    def read_moons(self, input_file) -> tuple:
        self.stdout.write(f"Importing moons from: {input_file} ...")
        moons = dict()
        ore_types = set()
        with input_file.open("r", encoding="utf-8") as fp:
            csv_reader = csv.DictReader(fp)
            for row in csv_reader:
                moon_id = int(row["moon_id"])
                ore_type_id = int(row["ore_type_id"])
                amount = float(row["amount"])
                if moon_id not in moons:
                    moons[moon_id] = list()
                moons[moon_id].append((ore_type_id, amount))
                ore_types.add(ore_type_id)

        if not len(moons.keys()):
            raise CommandError("Import file contains no moons.")

        self.stdout.write(
            f"Input file contains {len(moons.keys())} moons.",
        )
        return moons, ore_types

    def fetch_missing_eve_objects(
        self, EveModel: type, ids_incoming: set, force_refetch: bool
    ):
        if force_refetch:
            ids_to_fetch = set(ids_incoming)
        else:
            ids_existing = set(EveModel.objects.values_list("id", flat=True))
            ids_to_fetch = set(ids_incoming) - ids_existing
        if not len(ids_to_fetch):
            logger.debug("No %s objects to fetch from ESI", EveModel.__name__)
            return
        self.stdout.write(
            f"Fetching {len(ids_to_fetch):,} {EveModel.__name__} objects from ESI",
            ending="",
        )
        for id in ids_to_fetch:
            self._fetch_eve_object(EveModel, id)
        self.stdout.write("")
        ids_existing = set(EveModel.objects.values_list("id", flat=True))
        ids_missing = ids_to_fetch - ids_existing
        if ids_missing:
            logger.debug(
                "%s: Missing %d ids: %s",
                EveModel.__name__,
                len(ids_missing),
                ids_missing,
            )
            raise CommandError(
                f"Failed to fetch all {EveModel.__name__} objects. Please try again"
            )

    def _fetch_eve_object(self, EveModel: type, id: int):
        for run in range(MAX_RETRIES + 1):
            try:
                with transaction.atomic():
                    logger.info(
                        "Fetching %s object with id %s from ESI... %s",
                        EveModel.__name__,
                        id,
                        f"(retry #{run + 1})" if run > 0 else "",
                    )
                    EveModel.objects.update_or_create_esi(id=id)
            except (
                HTTPBadGateway,
                HTTPGatewayTimeout,
                HTTPServiceUnavailable,
                IntegrityError,
            ) as ex:
                logger.exception("Recoverable error occurred: %s", ex)
            else:
                self.stdout.write(".", ending="")
                break

    @transaction.atomic()
    def import_moons(self, moons, force_update):
        ids_incoming = set(moons.keys())
        ids_existing = set(Moon.objects.values_list("pk", flat=True))
        ids_missing = ids_incoming - ids_existing
        new_moons = {
            moon_id: moon for moon_id, moon in moons.items() if moon_id in ids_missing
        }
        if not len(new_moons):
            logger.debug("No Moon objects to create")
            if not force_update:
                self.stdout.write("Moons already exist in DB. Aborting.")
                return
        else:
            self.stdout.write(
                f"Creating {len(new_moons):,} moons...",
            )
            moon_objects = [Moon(eve_moon_id=moon_id) for moon_id in new_moons.keys()]
            Moon.objects.bulk_create(moon_objects, batch_size=BULK_BATCH_SIZE)
        if force_update:
            ids_to_update = ids_incoming - ids_missing
            self.stdout.write(
                f"Updating {len(ids_to_update):,} moons...",
            )
            Moon.objects.filter(pk__in=ids_to_update).update(
                products_updated_at=None, products_updated_by=None
            )
            MoonProduct.objects.filter(moon__pk__in=moons.keys()).delete()
            product_moons = moons
        else:
            product_moons = new_moons

        # create moon products
        if len(product_moons):
            self.stdout.write(
                f"Writing approx. {len(product_moons) * 3.5:,.0f} moon products..."
            )
            product_objects = []
            for moon_id, moon_data in product_moons.items():
                for ore in moon_data:
                    product_objects.append(
                        MoonProduct(moon_id=moon_id, ore_type_id=ore[0], amount=ore[1])
                    )
            MoonProduct.objects.bulk_create(product_objects, batch_size=BULK_BATCH_SIZE)

    def update_moons(self, moons):
        moons_to_update_qs = Moon.objects.filter(pk__in=moons.keys())
        for moon in moons_to_update_qs:
            tasks.update_moon_calculated_properties.delay(moon.pk)
        self.stdout.write(
            f"Updating calculated properties for {moons_to_update_qs.count():,} "
            "moons started..."
        )



from django.test import TestCase

from app_utils.testdata_factories import UserFactory

from .testdata.load_eveuniverse import load_eveuniverse


class TestAdminUI(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eveuniverse()

    def test_should_open_prices_page(self):
        # def
        user = UserFactory(is_superuser=True, is_staff=True)
        self.client.force_login(user)
        # when
        response = self.client.get("/admin/moonmining/eveoretype/")
        # then
        self.assertEqual(response.status_code, 200)



import datetime as dt
from typing import Optional

from django import template

from ..constants import DATETIME_FORMAT

register = template.Library()


@register.filter
def formatisk(value, magnitude: str = None) -> Optional[str]:
    """Return the formatted ISK value or None if input was invalid.

    Args:
    - magnitude: use the given magnitude to format the number, e.g. "b"
    """
    try:
        value = float(value)
    except (ValueError, TypeError):
        return None
    power_map = {"t": 12, "b": 9, "m": 6, "k": 3, "": 0}
    if magnitude not in power_map:
        if value >= 10**12:
            magnitude = "t"
        elif value >= 10**9:
            magnitude = "b"
        elif value >= 10**6:
            magnitude = "m"
        elif value >= 10**3:
            magnitude = "k"
        else:
            magnitude = ""
    return f"{value / 10 ** power_map[magnitude]:,.1f}{magnitude}"


@register.filter
def datetime(value: dt.datetime) -> Optional[str]:
    try:
        return value.strftime(DATETIME_FORMAT)
    except AttributeError:
        return None



import inspect
import os

_currentdir = os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))


def fetch_survey_data() -> str:
    survey_data = dict()
    with open(f"{_currentdir}/moon_survey_input_2.txt", "r", encoding="utf-8") as f:
        survey_data[2] = f.read()

    with open(f"{_currentdir}/moon_survey_input_3.txt", "r", encoding="utf-8") as f:
        survey_data[3] = f.read()

    return survey_data



# Generated by Django 4.0.3 on 2022-04-22 18:19

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("moonmining", "0005_make_pricing_more_flexible"),
    ]

    operations = [
        migrations.CreateModel(
            name="Label",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("description", models.TextField(blank=True, default="")),
                ("name", models.CharField(max_length=100, unique=True)),
                (
                    "style",
                    models.CharField(
                        choices=[
                            ("primary", "dark blue"),
                            ("success", "green"),
                            ("default", "grey"),
                            ("info", "light blue"),
                            ("warning", "orange"),
                            ("danger", "red"),
                        ],
                        default="default",
                        max_length=16,
                    ),
                ),
            ],
        ),
        migrations.AddField(
            model_name="moon",
            name="label",
            field=models.ForeignKey(
                default=None,
                null=True,
                on_delete=django.db.models.deletion.SET_DEFAULT,
                to="moonmining.label",
            ),
        ),
    ]



from django.utils.translation import gettext_lazy as _

from allianceauth import hooks
from allianceauth.services.hooks import MenuItemHook, UrlHook

from . import __title__, urls


class MoonMenu(MenuItemHook):
    def __init__(self):
        MenuItemHook.__init__(
            self,
            _(__title__),
            "fas fa-moon fa-fw",
            "moonmining:index",
            navactive=["moonmining:"],
        )

    def render(self, request):
        if request.user.has_perm("moonmining.basic_access"):
            return MenuItemHook.render(self, request)
        return ""


@hooks.register("menu_item_hook")
def register_menu():
    return MoonMenu()


@hooks.register("url_hook")
def register_url():
    return UrlHook(urls, "moonmining", r"^moonmining/")



from django.urls import path

from . import views

app_name = "moonmining"

urlpatterns = [
    path("", views.index, name="index"),
    path("add_owner", views.add_owner, name="add_owner"),
    path("upload_survey", views.upload_survey, name="upload_survey"),
    path("modal_loader_body", views.modal_loader_body, name="modal_loader_body"),
    path("tests", views.tests, name="tests"),
    # extractions
    path("extractions", views.extractions, name="extractions"),
    path(
        "extractions_data/<str:category>",
        views.extractions_data,
        name="extractions_data",
    ),
    path(
        "extraction/<int:extraction_pk>",
        views.extraction_details,
        name="extraction_details",
    ),
    path(
        "extraction_ledger/<int:extraction_pk>",
        views.extraction_ledger,
        name="extraction_ledger",
    ),
    # moons
    path("moons", views.moons, name="moons"),
    # path("moons_data/<str:category>", views.moons_data, name="moons_data"),
    path("moons_data/<str:category>", views.MoonListJson.as_view(), name="moons_data"),
    path("moons_fdd_data/<str:category>", views.moons_fdd_data, name="moons_fdd_data"),
    path("moon/<int:moon_pk>", views.moon_details, name="moon_details"),
    # reports
    path("reports", views.reports, name="reports"),
    path(
        "report_owned_value_data",
        views.report_owned_value_data,
        name="report_owned_value_data",
    ),
    path(
        "report_user_mining_data",
        views.report_user_mining_data,
        name="report_user_mining_data",
    ),
    path(
        "report_user_uploaded_data",
        views.report_user_uploaded_data,
        name="report_user_uploaded_data",
    ),
    path(
        "report_ore_prices_data",
        views.report_ore_prices_data,
        name="report_ore_prices_data",
    ),
]



import datetime as dt
from unittest.mock import Mock, patch

import pytz

from django.contrib.sessions.middleware import SessionMiddleware
from django.test import RequestFactory, TestCase
from django.urls import reverse
from django.utils.timezone import now
from esi.models import Token
from eveuniverse.models import EveMarketPrice, EveMoon

from allianceauth.eveonline.models import EveCorporationInfo
from app_utils.testing import (
    create_user_from_evecharacter,
    json_response_to_dict,
    json_response_to_python,
)

from .. import views
from ..models import EveOreType, Extraction, Label, Moon, Owner
from . import helpers
from .testdata.factories import (
    EveEntityCharacterFactory,
    EveEntityCorporationFactory,
    ExtractionFactory,
    MiningLedgerRecordFactory,
    MoonFactory,
    RefineryFactory,
)
from .testdata.load_allianceauth import load_allianceauth
from .testdata.load_eveuniverse import load_eveuniverse

MODULE_PATH = "moonmining.views"


class TestOwner(TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        load_allianceauth()
        load_eveuniverse()
        cls.factory = RequestFactory()
        cls.user, cls.character_ownership = create_user_from_evecharacter(
            1001, permissions=["moonmining.add_refinery_owner"]
        )

    @patch(MODULE_PATH + ".notify_admins")
    @patch(MODULE_PATH + ".tasks.update_owner")
    @patch(MODULE_PATH + ".messages_plus")
    def test_should_add_new_owner(
        self, mock_messages, mock_update_owner, mock_notify_admins
    ):
        # given
        token = Mock(spec=Token)
        token.character_id = self.character_ownership.character.character_id
        request = self.factory.get(reverse("moonmining:add_owner"))
        request.user = self.user
        request.token = token
        middleware = SessionMiddleware(Mock())
        middleware.process_request(request)
        orig_view = views.add_owner.__wrapped__.__wrapped__.__wrapped__
        # when
        response = orig_view(request, token)
        # then
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.url, reverse("moonmining:index"))
        self.assertTrue(mock_messages.success.called)
        self.assertTrue(mock_update_owner.delay.called)
        self.assertTrue(mock_notify_admins.called)
        obj = Owner.objects.get(corporation__corporation_id=2001)
        self.assertEqual(obj.character_ownership, self.character_ownership)

    @patch(MODULE_PATH + ".tasks.update_owner")
    @patch(MODULE_PATH + ".messages_plus")
    def test_should_update_existing_owner(self, mock_messages, mock_update_owner):
        # given
        Owner.objects.create(
            corporation=EveCorporationInfo.objects.get(corporation_id=2001),
            character_ownership=None,
        )
        token = Mock(spec=Token)
        token.character_id = self.character_ownership.character.character_id
        request = self.factory.get(reverse("moonmining:add_owner"))
        request.user = self.user
        request.token = token
        middleware = SessionMiddleware(Mock())
        middleware.process_request(request)
        orig_view = views.add_owner.__wrapped__.__wrapped__.__wrapped__
        # when
        response = orig_view(request, token)
        # then
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.url, reverse("moonmining:index"))
        self.assertTrue(mock_messages.success.called)
        self.assertTrue(mock_update_owner.delay.called)
        obj = Owner.objects.get(corporation__corporation_id=2001)
        self.assertEqual(obj.character_ownership, self.character_ownership)

    @patch(MODULE_PATH + ".tasks.update_owner")
    @patch(MODULE_PATH + ".messages_plus")
    def test_should_raise_404_if_character_ownership_not_found(
        self, mock_messages, mock_update_owner
    ):
        # given
        token = Mock(spec=Token)
        token.character_id = 1099
        request = self.factory.get(reverse("moonmining:add_owner"))
        request.user = self.user
        request.token = token
        middleware = SessionMiddleware(Mock())
        middleware.process_request(request)
        orig_view = views.add_owner.__wrapped__.__wrapped__.__wrapped__
        # when
        response = orig_view(request, token)
        # then
        self.assertEqual(response.status_code, 404)


class TestMoonsData(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eveuniverse()
        load_allianceauth()
        helpers.generate_market_prices()
        cls.moon = MoonFactory(eve_moon=EveMoon.objects.get(id=40161708))
        cls.moon.label = Label.objects.create(name="Dummy")
        cls.moon.save()
        MoonFactory(eve_moon=EveMoon.objects.get(id=40131695))
        MoonFactory(eve_moon=EveMoon.objects.get(id=40161709))

    @staticmethod
    def _response_to_dict(response):
        data = helpers.json_response_to_python_2(response)
        return {int(obj[0]): obj for obj in data}

    def test_should_return_all_moons(self):
        # given
        user, _ = create_user_from_evecharacter(
            1001,
            permissions=[
                "moonmining.basic_access",
                "moonmining.extractions_access",
                "moonmining.view_all_moons",
            ],
            scopes=Owner.esi_scopes(),
        )
        self.client.force_login(user)
        # when
        response = self.client.get(f"/moonmining/moons_data/{views.MoonsCategory.ALL}")
        # then
        self.assertEqual(response.status_code, 200)
        data = self._response_to_dict(response)
        self.assertSetEqual(set(data.keys()), {40131695, 40161708, 40161709})
        obj = data[40161708]
        self.assertEqual(obj[1], "Auga V - 1")

    def test_should_return_our_moons_only(self):
        # given
        moon = Moon.objects.get(pk=40131695)
        RefineryFactory(moon=moon)
        user, _ = create_user_from_evecharacter(
            1002,
            permissions=["moonmining.basic_access", "moonmining.extractions_access"],
            scopes=Owner.esi_scopes(),
        )
        self.client.force_login(user)
        # when
        response = self.client.get(f"/moonmining/moons_data/{views.MoonsCategory.OURS}")
        # then
        self.assertEqual(response.status_code, 200)
        data = self._response_to_dict(response)
        self.assertSetEqual(set(data.keys()), {40131695})

    def test_should_handle_empty_refineries(self):
        # given
        user, _ = create_user_from_evecharacter(
            1002,
            permissions=["moonmining.basic_access", "moonmining.extractions_access"],
            scopes=Owner.esi_scopes(),
        )
        self.client.force_login(user)
        moon = Moon.objects.get(pk=40131695)
        refinery = RefineryFactory(moon=moon)
        RefineryFactory(owner=refinery.owner, moon=None)
        # when
        response = self.client.get(f"/moonmining/moons_data/{views.MoonsCategory.OURS}")
        # then
        self.assertEqual(response.status_code, 200)
        data = self._response_to_dict(response)
        self.assertSetEqual(set(data.keys()), {40131695})

    def test_should_return_empty_list_for_all_moons(self):
        # given
        user, _ = create_user_from_evecharacter(
            1001,
            permissions=["moonmining.basic_access", "moonmining.extractions_access"],
            scopes=Owner.esi_scopes(),
        )
        self.client.force_login(user)
        # when
        response = self.client.get(f"/moonmining/moons_data/{views.MoonsCategory.ALL}")
        # then
        self.assertEqual(response.status_code, 200)
        data = self._response_to_dict(response)
        self.assertEqual(len(data), 0)

    def test_should_return_empty_list_for_our_moons(self):
        # given
        user, _ = create_user_from_evecharacter(
            1001,
            permissions=["moonmining.basic_access"],
            scopes=Owner.esi_scopes(),
        )
        self.client.force_login(user)
        # when
        response = self.client.get(f"/moonmining/moons_data/{views.MoonsCategory.OURS}")
        # then
        self.assertEqual(response.status_code, 200)
        data = self._response_to_dict(response)
        self.assertEqual(len(data), 0)

    def test_should_return_uploaded_moons_only(self):
        # given
        user, _ = create_user_from_evecharacter(
            1001,
            permissions=["moonmining.basic_access", "moonmining.upload_moon_scan"],
            scopes=Owner.esi_scopes(),
        )
        self.client.force_login(user)
        self.moon.products_updated_by = user
        self.moon.save()
        # when
        response = self.client.get(
            f"/moonmining/moons_data/{views.MoonsCategory.UPLOADS}"
        )
        # then
        self.assertEqual(response.status_code, 200)
        data = self._response_to_dict(response)
        self.assertSetEqual(set(data.keys()), {40161708})

    def test_should_return_fdd_for_all_moons(self):
        # given
        user, _ = create_user_from_evecharacter(
            1002,
            permissions=[
                "moonmining.basic_access",
                "moonmining.extractions_access",
                "moonmining.view_all_moons",
            ],
            scopes=Owner.esi_scopes(),
        )
        moon = Moon.objects.get(pk=40131695)
        RefineryFactory(moon=moon)
        self.client.force_login(user)
        # when
        response = self.client.get(
            f"/moonmining/moons_fdd_data/{views.MoonsCategory.ALL}"
            "?columns=alliance_name,corporation_name,region_name,"
            "constellation_name,solar_system_name,rarity_class_str,label_name,"
            "has_refinery_str,has_extraction_str,invalid_column"
        )
        # then
        self.assertEqual(response.status_code, 200)
        data = json_response_to_python(response)
        self.assertListEqual(data["alliance_name"], ["Wayne Enterprises"])
        self.assertListEqual(data["corporation_name"], ["Wayne Technologies"])
        self.assertListEqual(data["region_name"], ["Heimatar", "Metropolis"])
        self.assertListEqual(data["constellation_name"], ["Aldodan", "Hed"])
        self.assertListEqual(data["solar_system_name"], ["Auga", "Helgatild"])
        self.assertListEqual(data["rarity_class_str"], ["R64"])
        self.assertListEqual(data["label_name"], ["Dummy"])
        self.assertListEqual(data["has_refinery_str"], ["no", "yes"])
        self.assertListEqual(data["has_extraction_str"], ["no"])
        self.assertIn("ERROR", data["invalid_column"][0])


class TestMoonInfo(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eveuniverse()
        load_allianceauth()

    def test_should_open_page(self):
        # given
        moon = MoonFactory()
        user, _ = create_user_from_evecharacter(
            1002,
            permissions=[
                "moonmining.basic_access",
                "moonmining.extractions_access",
                "moonmining.view_all_moons",
                "moonmining.upload_moon_scan",
                "moonmining.add_refinery_owner",
            ],
            scopes=[
                "esi-industry.read_corporation_mining.v1",
                "esi-universe.read_structures.v1",
                "esi-characters.read_notifications.v1",
                "esi-corporations.read_structures.v1",
            ],
        )
        self.client.force_login(user)
        # when
        response = self.client.get(f"/moonmining/moon/{moon.pk}")
        # then
        self.assertTemplateUsed(response, "moonmining/modals/moon_details.html")


class TestViewsAreWorking(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eveuniverse()
        load_allianceauth()
        cls.user, _ = create_user_from_evecharacter(
            1002,
            permissions=[
                "moonmining.basic_access",
                "moonmining.extractions_access",
                "moonmining.reports_access",
                "moonmining.view_all_moons",
                "moonmining.upload_moon_scan",
                "moonmining.add_refinery_owner",
            ],
            scopes=[
                "esi-industry.read_corporation_mining.v1",
                "esi-universe.read_structures.v1",
                "esi-characters.read_notifications.v1",
                "esi-corporations.read_structures.v1",
            ],
        )
        cls.moon = MoonFactory()

    def test_should_redirect_to_extractions_page(self):
        # given
        self.client.force_login(self.user)
        # when
        response = self.client.get("/moonmining/")
        # then
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.url, "/moonmining/extractions")

    def test_should_open_extractions_page(self):
        # given
        self.client.force_login(self.user)
        # when
        response = self.client.get("/moonmining/extractions")
        # then
        self.assertTemplateUsed(response, "moonmining/extractions.html")

    def test_should_open_moon_details_page(self):
        # given
        self.client.force_login(self.user)
        # when
        response = self.client.get(f"/moonmining/moon/{self.moon.pk}?new_page=yes")
        # then
        self.assertTemplateUsed(response, "moonmining/_generic_modal_page.html")

    def test_should_open_extraction_details_page(self):
        # given
        refinery = RefineryFactory(moon=self.moon)
        extraction = ExtractionFactory(refinery=refinery)
        self.client.force_login(self.user)
        # when
        response = self.client.get(
            f"/moonmining/extraction/{extraction.pk}?new_page=yes"
        )
        # then
        self.assertTemplateUsed(response, "moonmining/_generic_modal_page.html")

    def test_should_open_add_moon_scan_page(self):
        # given
        self.client.force_login(self.user)
        # when
        response = self.client.get("/moonmining/upload_survey")
        # then
        self.assertTemplateUsed(response, "moonmining/modals/upload_survey.html")

    def test_should_open_moons_page(self):
        # given
        self.client.force_login(self.user)
        # when
        response = self.client.get("/moonmining/moons")
        # then
        self.assertTemplateUsed(response, "moonmining/moons.html")

    def test_should_open_reports_page(self):
        # given
        self.client.force_login(self.user)
        # when
        response = self.client.get("/moonmining/reports")
        # then
        self.assertTemplateUsed(response, "moonmining/reports.html")

    def test_should_handle_empty_refineries_extractions_page(self):
        # given
        refinery = RefineryFactory()
        ExtractionFactory(refinery=refinery)
        self.client.force_login(self.user)
        # when
        response = self.client.get("/moonmining/extractions")
        # then
        self.assertTemplateUsed(response, "moonmining/extractions.html")


class TestExtractionsData(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eveuniverse()
        load_allianceauth()
        helpers.generate_eve_entities_from_allianceauth()
        moon = MoonFactory(eve_moon=EveMoon.objects.get(id=40161708))
        cls.refinery = RefineryFactory(moon=moon)
        cls.extraction = ExtractionFactory(
            refinery=cls.refinery,
            chunk_arrival_at=dt.datetime(2019, 11, 20, 0, 1, 0, tzinfo=pytz.UTC),
            auto_fracture_at=dt.datetime(2019, 11, 20, 3, 1, 0, tzinfo=pytz.UTC),
            started_by_id=1001,
            started_at=now() - dt.timedelta(days=3),
            status=Extraction.Status.COMPLETED,
        )
        EveMarketPrice.objects.create(eve_type_id=45506, average_price=10)
        cls.user_1003, _ = create_user_from_evecharacter(1003)

    def test_should_show_extraction(self):
        # given
        MiningLedgerRecordFactory(
            refinery=self.refinery,
            character_id=1001,
            day=dt.date(2019, 11, 20),
            corporation_id=2001,
            user=self.user_1003,
        )
        user, _ = create_user_from_evecharacter(
            1002,
            permissions=[
                "moonmining.basic_access",
                "moonmining.extractions_access",
                "moonmining.view_moon_ledgers",
            ],
            scopes=Owner.esi_scopes(),
        )
        self.client.force_login(user)
        # when
        response = self.client.get(
            f"/moonmining/extractions_data/{views.ExtractionsCategory.PAST}",
        )
        # then
        self.assertEqual(response.status_code, 200)
        data = json_response_to_dict(response)
        self.assertSetEqual(set(data.keys()), {self.extraction.pk})
        obj = data[self.extraction.pk]
        self.assertIn("2019-Nov-20 00:01", obj["chunk_arrival_at"]["display"])
        self.assertEqual(obj["corporation_name"], "Wayne Technologies [WYN]")
        self.assertIn("modalExtractionLedger", obj["details"])

    def test_should_not_show_extraction(self):
        # given
        user, _ = create_user_from_evecharacter(
            1002,
            permissions=["moonmining.basic_access"],
            scopes=Owner.esi_scopes(),
        )
        self.client.force_login(user)
        # when
        response = self.client.get(
            f"/moonmining/extractions_data/{views.ExtractionsCategory.PAST}",
        )
        # then
        self.assertEqual(response.status_code, 302)

    def test_should_not_show_ledger_button_wo_permission(self):
        # given
        MiningLedgerRecordFactory(
            refinery=self.refinery,
            character_id=1001,
            day=dt.date(2019, 11, 20),
            corporation_id=2001,
            user=self.user_1003,
        )
        user, _ = create_user_from_evecharacter(
            1002,
            permissions=["moonmining.basic_access", "moonmining.extractions_access"],
            scopes=Owner.esi_scopes(),
        )
        self.client.force_login(user)
        # when
        response = self.client.get(
            f"/moonmining/extractions_data/{views.ExtractionsCategory.PAST}",
        )
        # then
        self.assertEqual(response.status_code, 200)
        data = json_response_to_dict(response)
        obj = data[self.extraction.pk]
        self.assertNotIn("modalExtractionLedger", obj["details"])

    def test_should_not_show_ledger_button_when_no_data(self):
        # given
        user, _ = create_user_from_evecharacter(
            1002,
            permissions=[
                "moonmining.basic_access",
                "moonmining.extractions_access",
                "moonmining.view_moon_ledgers",
            ],
            scopes=Owner.esi_scopes(),
        )
        self.client.force_login(user)
        # when
        response = self.client.get(
            f"/moonmining/extractions_data/{views.ExtractionsCategory.PAST}",
        )
        # then
        self.assertEqual(response.status_code, 200)
        data = json_response_to_dict(response)
        obj = data[self.extraction.pk]
        self.assertNotIn("modalExtractionLedger", obj["details"])


class TestReportsData(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eveuniverse()
        load_allianceauth()
        helpers.generate_eve_entities_from_allianceauth()
        cls.moon = MoonFactory(eve_moon=EveMoon.objects.get(id=40161708))
        cls.refinery = RefineryFactory(moon=cls.moon)
        cls.user, _ = create_user_from_evecharacter(
            1002,
            permissions=["moonmining.basic_access", "moonmining.reports_access"],
            scopes=Owner.esi_scopes(),
        )
        MoonFactory(
            eve_moon=EveMoon.objects.get(id=40131695), products_updated_by=cls.user
        )
        MoonFactory(
            eve_moon=EveMoon.objects.get(id=40161709), products_updated_by=cls.user
        )

    def test_should_return_owned_moon_values(self):
        # given
        self.client.force_login(self.user)
        # when
        response = self.client.get("/moonmining/report_owned_value_data")
        # then
        self.assertEqual(response.status_code, 200)
        # TODO: Test values

    def test_should_return_user_mining_data(self):
        # given
        today = dt.datetime(2021, 1, 15, 12, 0, tzinfo=pytz.UTC)
        months_1 = dt.datetime(2020, 12, 15, 12, 0, tzinfo=pytz.UTC)
        months_2 = dt.datetime(2020, 11, 15, 12, 0, tzinfo=pytz.UTC)
        months_3 = dt.datetime(2020, 10, 15, 12, 0, tzinfo=pytz.UTC)
        EveMarketPrice.objects.create(eve_type_id=45506, average_price=10)
        EveMarketPrice.objects.create(eve_type_id=45494, average_price=20)
        EveOreType.objects.update_current_prices(use_process_pricing=False)
        character = EveEntityCharacterFactory()
        corporation = EveEntityCorporationFactory()
        MiningLedgerRecordFactory(
            refinery=self.refinery,
            day=today.date() - dt.timedelta(days=1),
            character=character,
            corporation=corporation,
            ore_type_id=45506,
            quantity=100,
            user=self.user,
        )
        MiningLedgerRecordFactory(
            refinery=self.refinery,
            day=today.date() - dt.timedelta(days=2),
            character=character,
            corporation=corporation,
            ore_type_id=45494,
            quantity=200,
            user=self.user,
        )
        MiningLedgerRecordFactory(
            refinery=self.refinery,
            day=months_1.date() - dt.timedelta(days=1),
            character=character,
            corporation=corporation,
            ore_type_id=45494,
            quantity=200,
            user=self.user,
        )
        MiningLedgerRecordFactory(
            refinery=self.refinery,
            day=months_2.date() - dt.timedelta(days=1),
            character=character,
            corporation=corporation,
            ore_type_id=45494,
            quantity=500,
            user=self.user,
        )
        MiningLedgerRecordFactory(
            refinery=self.refinery,
            day=months_3.date() - dt.timedelta(days=1),
            character=character,
            corporation=corporation,
            ore_type_id=45494,
            quantity=600,
            user=self.user,
        )
        self.client.force_login(self.user)
        # when
        with patch(MODULE_PATH + ".now") as mock_now:
            mock_now.return_value = today
            response = self.client.get("/moonmining/report_user_mining_data")
        # then
        self.assertEqual(response.status_code, 200)
        data = json_response_to_dict(response)
        row = data[self.user.id]
        self.assertEqual(row["volume_month_0"], 100 * 10 + 200 * 10)
        self.assertEqual(row["price_month_0"], 10 * 100 + 20 * 200)
        self.assertEqual(row["volume_month_1"], 200 * 10)
        self.assertEqual(row["price_month_1"], 20 * 200)
        self.assertEqual(row["volume_month_2"], 500 * 10)
        self.assertEqual(row["price_month_2"], 20 * 500)
        self.assertEqual(row["volume_month_3"], 600 * 10)
        self.assertEqual(row["price_month_3"], 20 * 600)

    def test_should_return_user_uploads_data(self):
        # given
        self.client.force_login(self.user)
        # when
        response = self.client.get("/moonmining/report_user_uploaded_data")
        # then
        self.assertEqual(response.status_code, 200)
        user_data = [
            row
            for row in json_response_to_python(response)
            if row["name"] == self.user.profile.main_character.character_name
        ]
        self.assertEqual(user_data[0]["num_moons"], 2)

    def test_should_return_ore_prices(self):
        # given
        helpers.generate_market_prices()
        self.client.force_login(self.user)
        # when
        response = self.client.get("/moonmining/report_ore_prices_data")
        # then
        self.assertEqual(response.status_code, 200)
        data = json_response_to_dict(response)
        ore = data[45506]
        self.assertEqual(ore["name"], "Cinnabar")
        self.assertEqual(ore["price"], 2400.0)
        self.assertEqual(ore["group"], "Rare Moon Asteroids")
        self.assertEqual(ore["rarity_str"], "R32")


class TestExtractionLedgerData(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eveuniverse()
        load_allianceauth()
        helpers.generate_eve_entities_from_allianceauth()
        moon = MoonFactory(eve_moon=EveMoon.objects.get(id=40161708))
        cls.refinery = RefineryFactory(moon=moon)
        cls.extraction = ExtractionFactory(
            refinery=cls.refinery,
            chunk_arrival_at=dt.datetime(2019, 11, 20, 0, 1, 0, tzinfo=pytz.UTC),
            auto_fracture_at=dt.datetime(2019, 11, 20, 3, 1, 0, tzinfo=pytz.UTC),
            started_by_id=1001,
            started_at=now() - dt.timedelta(days=3),
            status=Extraction.Status.STARTED,
        )
        user_1003, _ = create_user_from_evecharacter(
            1003,
            permissions=[
                "moonmining.basic_access",
                "moonmining.extractions_access",
                "moonmining.view_moon_ledgers",
            ],
            scopes=Owner.esi_scopes(),
        )
        EveMarketPrice.objects.create(eve_type_id=45506, average_price=10)
        MiningLedgerRecordFactory(
            refinery=cls.refinery,
            character_id=1001,
            day=dt.date(2021, 4, 18),
            ore_type_id=45506,
            corporation_id=2001,
            quantity=100,
            user=user_1003,
        )

    def test_should_show_ledger(self):
        # given
        user_1002, _ = create_user_from_evecharacter(
            1002,
            permissions=[
                "moonmining.basic_access",
                "moonmining.extractions_access",
                "moonmining.view_moon_ledgers",
            ],
            scopes=Owner.esi_scopes(),
        )
        self.client.force_login(user_1002)
        # when
        response = self.client.get(
            f"/moonmining/extraction_ledger/{self.extraction.pk}",
        )
        # then
        self.assertTemplateUsed(response, "moonmining/modals/extraction_ledger.html")

    def test_should_not_show_ledger(self):
        # given
        user_1002, _ = create_user_from_evecharacter(
            1002,
            permissions=[
                "moonmining.basic_access",
                "moonmining.extractions_access",
            ],
            scopes=Owner.esi_scopes(),
        )
        self.client.force_login(user_1002)
        # when
        response = self.client.get(
            f"/moonmining/extraction_ledger/{self.extraction.pk}",
        )
        # then
        self.assertEqual(response.status_code, 302)



import json
from pathlib import Path

from app_utils.esi_testing import EsiClientStub, EsiEndpoint


def load_test_data():
    file_path = Path(__file__).parent / "esi.json"
    with file_path.open("r", encoding="utf-8") as fp:
        return json.load(fp)


_esi_data = load_test_data()

_endpoints = [
    EsiEndpoint(
        "Character",
        "get_characters_character_id_notifications",
        "character_id",
    ),
    EsiEndpoint(
        "Corporation",
        "get_corporations_corporation_id_structures",
        "corporation_id",
        needs_token=True,
    ),
    EsiEndpoint(
        "Industry",
        "get_corporation_corporation_id_mining_extractions",
        "corporation_id",
        needs_token=True,
    ),
    EsiEndpoint(
        "Industry",
        "get_corporation_corporation_id_mining_observers",
        "corporation_id",
        needs_token=True,
    ),
    EsiEndpoint(
        "Industry",
        "get_corporation_corporation_id_mining_observers_observer_id",
        ("corporation_id", "observer_id"),
        needs_token=True,
    ),
    EsiEndpoint(
        "Universe",
        "get_universe_structures_structure_id",
        "structure_id",
        needs_token=True,
    ),
]

esi_client_stub = EsiClientStub(_esi_data, endpoints=_endpoints)
esi_client_error_stub = EsiClientStub(_esi_data, endpoints=_endpoints, http_error=True)



import datetime as dt

from django.http import HttpResponse
from eveuniverse.models import EveEntity


class EnumToDict:
    """Adds ability to an Enum class to be converted to a ordinary dict.

    This e.g. allows using Enums in Django templates.
    """

    @classmethod
    def to_dict(cls) -> dict:
        """Convert this enum to dict."""
        return {k: elem.value for k, elem in cls.__members__.items()}


class HttpResponseUnauthorized(HttpResponse):
    status_code = 401


def eveentity_get_or_create_esi_safe(id):
    """Get or Create EveEntity with given ID safely and return it. Else return None."""
    if id:
        try:
            entity, _ = EveEntity.objects.get_or_create_esi(id=id)
            return entity
        except OSError:
            pass
    return None


def round_seconds(obj: dt.datetime) -> dt.datetime:
    """Return copy rounded to full seconds."""
    if obj.microsecond >= 500_000:
        obj += dt.timedelta(seconds=1)
    return obj.replace(microsecond=0)



import datetime as dt
from unittest.mock import patch

import pytz

from django.test import TestCase, override_settings
from django.urls import reverse
from django.utils.timezone import now
from django_webtest import WebTest
from eveuniverse.models import EveMoon

from app_utils.esi import EsiStatus
from app_utils.testing import NoSocketsTestCase, create_user_from_evecharacter

from .. import tasks
from ..models import Owner, Refinery
from . import helpers
from .testdata.esi_client_stub import esi_client_stub
from .testdata.factories import (
    ExtractionFactory,
    MoonFactory,
    OwnerFactory,
    RefineryFactory,
)
from .testdata.load_allianceauth import load_allianceauth
from .testdata.load_eveuniverse import load_eveuniverse, nearest_celestial_stub
from .testdata.survey_data import fetch_survey_data

MANAGERS_PATH = "moonmining.managers"
MODELS_PATH = "moonmining.models"
TASKS_PATH = "moonmining.tasks"
VIEWS_PATH = "moonmining.views"


class TestUI(WebTest):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eveuniverse()
        load_allianceauth()
        cls.user, cls.character_ownership = create_user_from_evecharacter(
            1001,
            permissions=["moonmining.basic_access", "moonmining.extractions_access"],
        )

    def test_should_open_extractions(self):
        # given
        self.app.set_user(self.user)
        # when
        index = self.app.get(reverse("moonmining:extractions"))
        # then
        self.assertEqual(index.status_code, 200)

    # TODO: Add more UI tests


@patch(TASKS_PATH + ".fetch_esi_status", lambda: EsiStatus(True, 100, 60))
@patch(MODELS_PATH + ".EveSolarSystem.nearest_celestial", new=nearest_celestial_stub)
@override_settings(CELERY_ALWAYS_EAGER=True, CELERY_EAGER_PROPAGATES_EXCEPTIONS=True)
class TestUpdateTasks(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eveuniverse()
        load_allianceauth()
        helpers.generate_eve_entities_from_allianceauth()
        helpers.generate_market_prices()
        _, cls.character_ownership = helpers.create_default_user_from_evecharacter(1001)

    @patch(MODELS_PATH + ".esi")
    def test_should_update_all_mining_corporations(self, mock_esi):
        # given
        mock_esi.client = esi_client_stub
        MoonFactory(eve_moon=EveMoon.objects.get(id=40161708))
        corporation_2001 = OwnerFactory(character_ownership=self.character_ownership)
        # when
        tasks.run_regular_updates.delay()
        # then
        self.assertSetEqual(Refinery.objects.ids(), {1000000000001, 1000000000002})
        refinery = Refinery.objects.get(id=1000000000001)
        self.assertEqual(refinery.extractions.count(), 1)
        corporation_2001.refresh_from_db()
        self.assertAlmostEqual(
            corporation_2001.last_update_at, now(), delta=dt.timedelta(minutes=1)
        )
        self.assertTrue(corporation_2001.last_update_ok)

    @patch(MODELS_PATH + ".esi")
    def test_should_report_when_updating_mining_corporations_failed(self, mock_esi):
        # given
        mock_esi.client.Corporation.get_corporations_corporation_id_structures.side_effect = (
            OSError
        )
        corporation_2001 = OwnerFactory(character_ownership=self.character_ownership)
        # when
        try:
            tasks.run_regular_updates.delay()
        except OSError:
            pass
        # then
        corporation_2001.refresh_from_db()
        self.assertAlmostEqual(
            corporation_2001.last_update_at, now(), delta=dt.timedelta(minutes=1)
        )
        self.assertIsNone(corporation_2001.last_update_ok)

        # TODO: add more tests

    @patch(MODELS_PATH + ".esi")
    def test_should_not_update_disabled_corporation(self, mock_esi):
        # given
        mock_esi.client = esi_client_stub
        MoonFactory(eve_moon=EveMoon.objects.get(id=40161708))
        corporation_2001 = OwnerFactory(character_ownership=self.character_ownership)
        _, character_ownership_1003 = create_user_from_evecharacter(
            1003,
            permissions=[
                "moonmining.basic_access",
                "moonmining.extractions_access",
                "moonmining.add_refinery_owner",
            ],
            scopes=Owner.esi_scopes(),
        )
        corporation_2002 = OwnerFactory(
            character_ownership=character_ownership_1003, last_update_ok=None
        )
        my_date = dt.datetime(2020, 1, 11, 12, 30, tzinfo=pytz.UTC)
        corporation_2002.last_update_at = my_date
        corporation_2002.is_enabled = False
        corporation_2002.save()
        # when
        tasks.run_regular_updates.delay()
        # then
        corporation_2001.refresh_from_db()
        self.assertAlmostEqual(
            corporation_2001.last_update_at, now(), delta=dt.timedelta(minutes=1)
        )
        self.assertTrue(corporation_2001.last_update_ok)
        corporation_2002.refresh_from_db()
        self.assertEqual(corporation_2002.last_update_at, my_date)
        self.assertIsNone(corporation_2002.last_update_ok)

    @patch(MODELS_PATH + ".esi")
    def test_should_update_mining_ledgers(self, mock_esi):
        # given
        mock_esi.client = esi_client_stub
        owner_2001 = OwnerFactory(character_ownership=self.character_ownership)
        refinery_1 = RefineryFactory(id=1000000000001, owner=owner_2001)
        refinery_2 = RefineryFactory(id=1000000000002, owner=owner_2001)
        _, ownership_1003 = helpers.create_default_user_from_evecharacter(1003)
        owner_2002 = OwnerFactory(character_ownership=ownership_1003)
        refinery_11 = RefineryFactory.create(id=1000000000011, owner=owner_2002)
        # when
        tasks.run_report_updates()
        # then
        self.assertEqual(refinery_1.mining_ledger.count(), 2)
        self.assertEqual(refinery_2.mining_ledger.count(), 1)
        self.assertEqual(refinery_11.mining_ledger.count(), 1)

    @patch(TASKS_PATH + ".EveMarketPrice.objects.update_from_esi")
    def test_should_update_all_calculated_values(self, mock_update_prices):
        # given
        mock_update_prices.return_value = None
        moon = MoonFactory()
        owner = OwnerFactory(character_ownership=self.character_ownership)
        refinery = RefineryFactory(moon=moon, owner=owner)
        extraction = ExtractionFactory(refinery=refinery)
        # when
        tasks.run_calculated_properties_update.delay()
        # then
        moon.refresh_from_db()
        extraction.refresh_from_db()
        self.assertIsNotNone(moon.value)
        self.assertIsNotNone(extraction.value)
        ore = extraction.products.first().ore_type
        self.assertIsNotNone(ore.extras.current_price)


class TestProcessSurveyInput(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eveuniverse()
        load_allianceauth()
        cls.user, cls.character_ownership = create_user_from_evecharacter(
            1001,
            permissions=[
                "moonmining.basic_access",
                "moonmining.extractions_access",
                "moonmining.add_refinery_owner",
            ],
            scopes=Owner.esi_scopes(),
        )
        cls.survey_data = fetch_survey_data()

    @patch(MANAGERS_PATH + ".notify", new=lambda *args, **kwargs: None)
    def test_should_handle_bad_data_orderly(self):
        # when
        result = tasks.process_survey_input(self.survey_data.get(3))
        # then
        self.assertFalse(result)

    @patch(MANAGERS_PATH + ".notify")
    def test_notification_on_success(self, mock_notify):
        result = tasks.process_survey_input(self.survey_data.get(2), self.user.pk)
        self.assertTrue(result)
        self.assertTrue(mock_notify.called)
        _, kwargs = mock_notify.call_args
        self.assertEqual(kwargs["user"], self.user)
        self.assertEqual(kwargs["level"], "success")

    @patch(MANAGERS_PATH + ".notify")
    def test_notification_on_error_1(self, mock_notify):
        result = tasks.process_survey_input("invalid input", self.user.pk)
        self.assertFalse(result)
        self.assertTrue(mock_notify.called)
        _, kwargs = mock_notify.call_args
        self.assertEqual(kwargs["user"], self.user)
        self.assertEqual(kwargs["level"], "danger")



from django.apps import AppConfig

from . import __version__


class MoonPlanerConfig(AppConfig):
    name = "moonmining"
    label = "moonmining"
    verbose_name = "Moon Mining v{}".format(__version__)



# Generated by Django 3.2.9 on 2021-11-22 21:34

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("moonmining", "0003_mining_ledger_reports"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="general",
            options={
                "default_permissions": (),
                "managed": False,
                "permissions": (
                    ("basic_access", "Can access the moonmining app"),
                    (
                        "extractions_access",
                        "Can access extractions and view owned moons",
                    ),
                    ("reports_access", "Can access reports"),
                    ("view_all_moons", "Can view all known moons"),
                    ("upload_moon_scan", "Can upload moon scans"),
                    ("add_refinery_owner", "Can add refinery owner"),
                    ("view_moon_ledgers", "Can view moon ledgers"),
                ),
            },
        ),
    ]



# Generated by Django 3.1.6 on 2021-04-15 13:25

import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ("moonmining", "0002_add_mining_ledger"),
    ]

    operations = [
        migrations.AddField(
            model_name="miningledgerrecord",
            name="user",
            field=models.ForeignKey(
                default=None,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="mining_ledger",
                to=settings.AUTH_USER_MODEL,
            ),
        ),
        migrations.AddField(
            model_name="refinery",
            name="ledger_last_update_at",
            field=models.DateTimeField(
                default=None,
                help_text="last successful update of mining ledger",
                null=True,
            ),
        ),
        migrations.AddField(
            model_name="refinery",
            name="ledger_last_update_ok",
            field=models.BooleanField(
                default=None,
                help_text="True if the last update of the mining ledger was successful",
                null=True,
            ),
        ),
        migrations.CreateModel(
            name="EveOreTypeExtras",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "refined_price",
                    models.FloatField(
                        default=None,
                        help_text="price calculated as sum of prices for refined materials",
                        null=True,
                    ),
                ),
                (
                    "ore_type",
                    models.OneToOneField(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="extras",
                        to="moonmining.eveoretype",
                    ),
                ),
            ],
        ),
    ]



import datetime as dt

from django.test import TestCase

from .. import helpers


class TestRoundDatetime(TestCase):
    def test_should_round_up(self):
        # given
        my_dt = dt.datetime(2021, 4, 6, 20, 15, 0, 567000)
        # when/then
        self.assertEqual(
            dt.datetime(2021, 4, 6, 20, 15, 1), helpers.round_seconds(my_dt)
        )

    def test_should_round_down(self):
        # given
        my_dt = dt.datetime(2021, 4, 6, 20, 15, 0, 267000)
        # when/then
        self.assertEqual(
            dt.datetime(2021, 4, 6, 20, 15, 0), helpers.round_seconds(my_dt)
        )

    def test_should_do_nothing(self):
        # given
        my_dt = dt.datetime(2021, 4, 6, 20, 15, 0)
        # when/then
        self.assertEqual(
            dt.datetime(2021, 4, 6, 20, 15, 0), helpers.round_seconds(my_dt)
        )



# flake8: noqa
"""scripts generates moons for load testing

This script can be executed directly from shell.
"""

import os
import sys
from pathlib import Path

myauth_dir = Path(__file__).parent.parent.parent.parent.parent / "myauth"
sys.path.insert(0, str(myauth_dir))

import django
from django.apps import apps

# init and setup django project
os.environ.setdefault("DJANGO_SETTINGS_MODULE", "myauth.settings.local")
django.setup()

"""SCRIPT"""
import concurrent.futures
import csv
import random
from pathlib import Path

from django.contrib.auth.models import User
from django.utils.timezone import now
from eveuniverse.core.esitools import is_esi_online
from eveuniverse.models import EveMoon

from moonmining.models import EveOreType, Moon, MoonProduct

from .factories import random_percentages

MAX_MOONS = 1000

ORE_TYPES = {
    45490: "Zeolites",
    45491: "Sylvite",
    45492: "Bitumens",
    45493: "Coesite",
    45494: "Cobaltite",
    45495: "Euxenite",
    45496: "Titanite",
    45497: "Scheelite",
    45498: "Otavite",
    45499: "Sperrylite",
    45500: "Vanadinite",
    45501: "Chromite",
    45502: "Carnotite",
    45503: "Zircon",
    45504: "Pollucite",
    45506: "Cinnabar",
    45510: "Xenotime",
    45511: "Monazite",
    45512: "Loparite",
    45513: "Ytterbite",
}


def create_moon(moon_id):
    eve_moon, _ = EveMoon.objects.get_or_create_esi(id=moon_id)
    moon, created = Moon.objects.get_or_create(
        eve_moon=eve_moon,
        defaults={
            "products_updated_at": now(),
            "products_updated_by": random_user,
        },
    )
    if created:
        percentages = random_percentages(4)
        for ore_type_id in random.sample(ore_type_ids, k=4):
            ore_type, _ = EveOreType.objects.get_or_create_esi(id=ore_type_id)
            MoonProduct.objects.create(
                moon=moon, ore_type=ore_type, amount=percentages.pop()
            )
        moon.update_calculated_properties()


print(f"Generating {MAX_MOONS} moons...")
data_path = Path(__file__).parent / "moon_ids.csv"
with data_path.open("r", encoding="utf-8") as fp:
    csv_reader = csv.reader(fp)
    moon_ids = [int(obj[0]) for obj in list(csv_reader)]
ore_type_ids = list(ORE_TYPES.keys())
random_user = User.objects.order_by("?").first()

if not is_esi_online():
    print("ESI is offline - aborting")

with concurrent.futures.ThreadPoolExecutor(max_workers=50) as executor:
    executor.map(create_moon, random.sample(moon_ids, k=MAX_MOONS))

print()
print(f"{MAX_MOONS} moons generated.")
